# vim: set ft=python ts=4 sw=4 expandtab:
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2014-2015 Kenneth J. Pronovici.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# Version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Copies of the GNU General Public License are available from
# the Free Software Foundation website, http://www.gnu.org/.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Official Cedar Backup Extensions
# Purpose  : "Store" type extension that writes data to Amazon S3.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################################
# Module documentation
########################################################################

"""
Store-type extension that writes data to Amazon S3.

This extension requires a new configuration section <amazons3> and is intended
to be run immediately after the standard stage action, replacing the standard
store action.  Aside from its own configuration, it requires the options and
staging configuration sections in the standard Cedar Backup configuration file.
Since it is intended to replace the store action, it does not rely on any store
configuration.

The underlying functionality relies on the U{AWS CLI interface
<http://aws.amazon.com/documentation/cli/>}.  Before you use this extension,
you need to set up your Amazon S3 account and configure the AWS CLI connection
per Amazon's documentation.  The extension assumes that the backup is being
executed as root, and switches over to the configured backup user to
communicate with AWS.  So, make sure you configure AWS CLI as the backup user
and not root.

You can optionally configure Cedar Backup to encrypt data before sending it
to S3.  To do that, provide a complete command line using the ``${input``} and
``${output``} variables to represent the original input file and the encrypted
output file.  This command will be executed as the backup user.

For instance, you can use something like this with GPG::

   /usr/bin/gpg -c --no-use-agent --batch --yes --passphrase-file /home/backup/.passphrase -o ${output} ${input}

The GPG mechanism depends on a strong passphrase for security.  One way to
generate a strong passphrase is using your system random number generator, i.e.::

   dd if=/dev/urandom count=20 bs=1 | xxd -ps

(See U{StackExchange <http://security.stackexchange.com/questions/14867/gpg-encryption-security>}
for more details about that advice.) If you decide to use encryption, make sure
you save off the passphrase in a safe place, so you can get at your backup data
later if you need to.  And obviously, make sure to set permissions on the
passphrase file so it can only be read by the backup user.

This extension was written for and tested on Linux.  It will throw an exception
if run on Windows.

:author: Kenneth J. Pronovici <pronovic@ieee.org>
"""

########################################################################
# Imported modules
########################################################################

import datetime
import json
import logging
import os
import shutil
import sys
import tempfile
from functools import total_ordering

from CedarBackup3.actions.constants import DIR_TIME_FORMAT, STAGE_INDICATOR
from CedarBackup3.actions.util import writeIndicatorFile
from CedarBackup3.config import ByteQuantity, addByteQuantityNode, readByteQuantity
from CedarBackup3.filesystem import BackupFileList, FilesystemList
from CedarBackup3.util import (
    UNIT_BYTES,
    changeOwnership,
    displayBytes,
    executeCommand,
    isRunningAsRoot,
    isStartOfWeek,
    pathJoin,
    resolveCommand,
)
from CedarBackup3.xmlutil import (
    addBooleanNode,
    addContainerNode,
    addStringNode,
    createInputDom,
    readBoolean,
    readFirstChild,
    readString,
)

########################################################################
# Module-wide constants and variables
########################################################################

logger = logging.getLogger("CedarBackup3.log.extend.amazons3")

SU_COMMAND = ["su"]
AWS_COMMAND = ["aws"]

STORE_INDICATOR = "cback.amazons3"


########################################################################
# AmazonS3Config class definition
########################################################################


@total_ordering
class AmazonS3Config:
    """
    Class representing Amazon S3 configuration.

    Amazon S3 configuration is used for storing backup data in Amazon's S3 cloud
    storage using the ``s3cmd`` tool.

    The following restrictions exist on data in this class:

       - The s3Bucket value must be a non-empty string
       - The encryptCommand value, if set, must be a non-empty string
       - The full backup size limit, if set, must be a ByteQuantity >= 0
       - The incremental backup size limit, if set, must be a ByteQuantity >= 0

    """

    def __init__(
        self, warnMidnite=None, s3Bucket=None, encryptCommand=None, fullBackupSizeLimit=None, incrementalBackupSizeLimit=None
    ):
        """
        Constructor for the ``AmazonS3Config`` class.

        Args:
           warnMidnite: Whether to generate warnings for crossing midnite
           s3Bucket: Name of the Amazon S3 bucket in which to store the data
           encryptCommand: Command used to encrypt backup data before upload to S3
           fullBackupSizeLimit: Maximum size of a full backup, a ByteQuantity
           incrementalBackupSizeLimit: Maximum size of an incremental backup, a ByteQuantity

        Raises:
           ValueError: If one of the values is invalid
        """
        self._warnMidnite = None
        self._s3Bucket = None
        self._encryptCommand = None
        self._fullBackupSizeLimit = None
        self._incrementalBackupSizeLimit = None
        self.warnMidnite = warnMidnite
        self.s3Bucket = s3Bucket
        self.encryptCommand = encryptCommand
        self.fullBackupSizeLimit = fullBackupSizeLimit
        self.incrementalBackupSizeLimit = incrementalBackupSizeLimit

    def __repr__(self):
        """
        Official string representation for class instance.
        """
        return "AmazonS3Config(%s, %s, %s, %s, %s)" % (
            self.warnMidnite,
            self.s3Bucket,
            self.encryptCommand,
            self.fullBackupSizeLimit,
            self.incrementalBackupSizeLimit,
        )

    def __str__(self):
        """
        Informal string representation for class instance.
        """
        return self.__repr__()

    def __eq__(self, other):
        """Equals operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        """Less-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) < 0

    def __gt__(self, other):
        """Greater-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) > 0

    def __cmp__(self, other):
        """
        Original Python 2 comparison operator.
        Args:
           other: Other object to compare to
        Returns:
            -1/0/1 depending on whether self is ``<``, ``=`` or ``>`` other
        """
        if other is None:
            return 1
        if self.warnMidnite != other.warnMidnite:
            if self.warnMidnite < other.warnMidnite:
                return -1
            else:
                return 1
        if self.s3Bucket != other.s3Bucket:
            if str(self.s3Bucket or "") < str(other.s3Bucket or ""):
                return -1
            else:
                return 1
        if self.encryptCommand != other.encryptCommand:
            if str(self.encryptCommand or "") < str(other.encryptCommand or ""):
                return -1
            else:
                return 1
        if self.fullBackupSizeLimit != other.fullBackupSizeLimit:
            if (self.fullBackupSizeLimit or ByteQuantity()) < (other.fullBackupSizeLimit or ByteQuantity()):
                return -1
            else:
                return 1
        if self.incrementalBackupSizeLimit != other.incrementalBackupSizeLimit:
            if (self.incrementalBackupSizeLimit or ByteQuantity()) < (other.incrementalBackupSizeLimit or ByteQuantity()):
                return -1
            else:
                return 1
        return 0

    def _setWarnMidnite(self, value):
        """
        Property target used to set the midnite warning flag.
        No validations, but we normalize the value to ``True`` or ``False``.
        """
        if value:
            self._warnMidnite = True
        else:
            self._warnMidnite = False

    def _getWarnMidnite(self):
        """
        Property target used to get the midnite warning flag.
        """
        return self._warnMidnite

    def _setS3Bucket(self, value):
        """
        Property target used to set the S3 bucket.
        """
        if value is not None:
            if len(value) < 1:
                raise ValueError("S3 bucket must be non-empty string.")
        self._s3Bucket = value

    def _getS3Bucket(self):
        """
        Property target used to get the S3 bucket.
        """
        return self._s3Bucket

    def _setEncryptCommand(self, value):
        """
        Property target used to set the encrypt command.
        """
        if value is not None:
            if len(value) < 1:
                raise ValueError("Encrypt command must be non-empty string.")
        self._encryptCommand = value

    def _getEncryptCommand(self):
        """
        Property target used to get the encrypt command.
        """
        return self._encryptCommand

    def _setFullBackupSizeLimit(self, value):
        """
        Property target used to set the full backup size limit.
        The value must be an integer >= 0.
        Raises:
           ValueError: If the value is not valid
        """
        if value is None:
            self._fullBackupSizeLimit = None
        else:
            if isinstance(value, ByteQuantity):
                self._fullBackupSizeLimit = value
            else:
                self._fullBackupSizeLimit = ByteQuantity(value, UNIT_BYTES)

    def _getFullBackupSizeLimit(self):
        """
        Property target used to get the full backup size limit.
        """
        return self._fullBackupSizeLimit

    def _setIncrementalBackupSizeLimit(self, value):
        """
        Property target used to set the incremental backup size limit.
        The value must be an integer >= 0.
        Raises:
           ValueError: If the value is not valid
        """
        if value is None:
            self._incrementalBackupSizeLimit = None
        else:
            if isinstance(value, ByteQuantity):
                self._incrementalBackupSizeLimit = value
            else:
                self._incrementalBackupSizeLimit = ByteQuantity(value, UNIT_BYTES)

    def _getIncrementalBackupSizeLimit(self):
        """
        Property target used to get the incremental backup size limit.
        """
        return self._incrementalBackupSizeLimit

    warnMidnite = property(_getWarnMidnite, _setWarnMidnite, None, "Whether to generate warnings for crossing midnite.")
    s3Bucket = property(_getS3Bucket, _setS3Bucket, None, doc="Amazon S3 Bucket in which to store data")
    encryptCommand = property(_getEncryptCommand, _setEncryptCommand, None, doc="Command used to encrypt data before upload to S3")
    fullBackupSizeLimit = property(
        _getFullBackupSizeLimit, _setFullBackupSizeLimit, None, doc="Maximum size of a full backup, as a ByteQuantity"
    )
    incrementalBackupSizeLimit = property(
        _getIncrementalBackupSizeLimit,
        _setIncrementalBackupSizeLimit,
        None,
        doc="Maximum size of an incremental backup, as a ByteQuantity",
    )


########################################################################
# LocalConfig class definition
########################################################################


@total_ordering
class LocalConfig:
    """
    Class representing this extension's configuration document.

    This is not a general-purpose configuration object like the main Cedar
    Backup configuration object.  Instead, it just knows how to parse and emit
    amazons3-specific configuration values.  Third parties who need to read and
    write configuration related to this extension should access it through the
    constructor, ``validate`` and ``addConfig`` methods.

    *Note:* Lists within this class are "unordered" for equality comparisons.

    """

    def __init__(self, xmlData=None, xmlPath=None, validate=True):
        """
        Initializes a configuration object.

        If you initialize the object without passing either ``xmlData`` or
        ``xmlPath`` then configuration will be empty and will be invalid until it
        is filled in properly.

        No reference to the original XML data or original path is saved off by
        this class.  Once the data has been parsed (successfully or not) this
        original information is discarded.

        Unless the ``validate`` argument is ``False``, the :any:`LocalConfig.validate`
        method will be called (with its default arguments) against configuration
        after successfully parsing any passed-in XML.  Keep in mind that even if
        ``validate`` is ``False``, it might not be possible to parse the passed-in
        XML document if lower-level validations fail.

        *Note:* It is strongly suggested that the ``validate`` option always be set
        to ``True`` (the default) unless there is a specific need to read in
        invalid configuration from disk.

        Args:
           xmlData (String data): XML data representing configuration
           xmlPath (Absolute path to a file on disk): Path to an XML file on disk
           validate (Boolean true/false): Validate the document after parsing it
        Raises:
           ValueError: If both ``xmlData`` and ``xmlPath`` are passed-in
           ValueError: If the XML data in ``xmlData`` or ``xmlPath`` cannot be parsed
           ValueError: If the parsed configuration document is not valid
        """
        self._amazons3 = None
        self.amazons3 = None
        if xmlData is not None and xmlPath is not None:
            raise ValueError("Use either xmlData or xmlPath, but not both.")
        if xmlData is not None:
            self._parseXmlData(xmlData)
            if validate:
                self.validate()
        elif xmlPath is not None:
            with open(xmlPath) as f:
                xmlData = f.read()
            self._parseXmlData(xmlData)
            if validate:
                self.validate()

    def __repr__(self):
        """
        Official string representation for class instance.
        """
        return "LocalConfig(%s)" % (self.amazons3)

    def __str__(self):
        """
        Informal string representation for class instance.
        """
        return self.__repr__()

    def __eq__(self, other):
        """Equals operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        """Less-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) < 0

    def __gt__(self, other):
        """Greater-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) > 0

    def __cmp__(self, other):
        """
        Original Python 2 comparison operator.
        Lists within this class are "unordered" for equality comparisons.
        Args:
           other: Other object to compare to
        Returns:
            -1/0/1 depending on whether self is ``<``, ``=`` or ``>`` other
        """
        if other is None:
            return 1
        if self.amazons3 != other.amazons3:
            if self.amazons3 < other.amazons3:
                return -1
            else:
                return 1
        return 0

    def _setAmazonS3(self, value):
        """
        Property target used to set the amazons3 configuration value.
        If not ``None``, the value must be a ``AmazonS3Config`` object.
        Raises:
           ValueError: If the value is not a ``AmazonS3Config``
        """
        if value is None:
            self._amazons3 = None
        else:
            if not isinstance(value, AmazonS3Config):
                raise ValueError("Value must be a ``AmazonS3Config`` object.")
            self._amazons3 = value

    def _getAmazonS3(self):
        """
        Property target used to get the amazons3 configuration value.
        """
        return self._amazons3

    amazons3 = property(_getAmazonS3, _setAmazonS3, None, "AmazonS3 configuration in terms of a ``AmazonS3Config`` object.")

    def validate(self):
        """
        Validates configuration represented by the object.

        AmazonS3 configuration must be filled in.  Within that, the s3Bucket target must be filled in

        Raises:
           ValueError: If one of the validations fails
        """
        if self.amazons3 is None:
            raise ValueError("AmazonS3 section is required.")
        if self.amazons3.s3Bucket is None:
            raise ValueError("AmazonS3 s3Bucket must be set.")

    def addConfig(self, xmlDom, parentNode):
        """
        Adds an <amazons3> configuration section as the next child of a parent.

        Third parties should use this function to write configuration related to
        this extension.

        We add the following fields to the document::

           warnMidnite                 //cb_config/amazons3/warn_midnite
           s3Bucket                    //cb_config/amazons3/s3_bucket
           encryptCommand              //cb_config/amazons3/encrypt
           fullBackupSizeLimit         //cb_config/amazons3/full_size_limit
           incrementalBackupSizeLimit  //cb_config/amazons3/incr_size_limit

        Args:
           xmlDom: DOM tree as from ``impl.createDocument()``
           parentNode: Parent that the section should be appended to
        """
        if self.amazons3 is not None:
            sectionNode = addContainerNode(xmlDom, parentNode, "amazons3")
            addBooleanNode(xmlDom, sectionNode, "warn_midnite", self.amazons3.warnMidnite)
            addStringNode(xmlDom, sectionNode, "s3_bucket", self.amazons3.s3Bucket)
            addStringNode(xmlDom, sectionNode, "encrypt", self.amazons3.encryptCommand)
            addByteQuantityNode(xmlDom, sectionNode, "full_size_limit", self.amazons3.fullBackupSizeLimit)
            addByteQuantityNode(xmlDom, sectionNode, "incr_size_limit", self.amazons3.incrementalBackupSizeLimit)

    def _parseXmlData(self, xmlData):
        """
        Internal method to parse an XML string into the object.

        This method parses the XML document into a DOM tree (``xmlDom``) and then
        calls a static method to parse the amazons3 configuration section.

        Args:
           xmlData (String data): XML data to be parsed
        Raises:
           ValueError: If the XML cannot be successfully parsed
        """
        (_, parentNode) = createInputDom(xmlData)
        self._amazons3 = LocalConfig._parseAmazonS3(parentNode)

    @staticmethod
    def _parseAmazonS3(parent):
        """
        Parses an amazons3 configuration section.

        We read the following individual fields::

           warnMidnite                 //cb_config/amazons3/warn_midnite
           s3Bucket                    //cb_config/amazons3/s3_bucket
           encryptCommand              //cb_config/amazons3/encrypt
           fullBackupSizeLimit         //cb_config/amazons3/full_size_limit
           incrementalBackupSizeLimit  //cb_config/amazons3/incr_size_limit

        Args:
           parent: Parent node to search beneath

        Returns:
            ``AmazonS3Config`` object or ``None`` if the section does not exist
        Raises:
           ValueError: If some filled-in value is invalid
        """
        amazons3 = None
        section = readFirstChild(parent, "amazons3")
        if section is not None:
            amazons3 = AmazonS3Config()
            amazons3.warnMidnite = readBoolean(section, "warn_midnite")
            amazons3.s3Bucket = readString(section, "s3_bucket")
            amazons3.encryptCommand = readString(section, "encrypt")
            amazons3.fullBackupSizeLimit = readByteQuantity(section, "full_size_limit")
            amazons3.incrementalBackupSizeLimit = readByteQuantity(section, "incr_size_limit")
        return amazons3


########################################################################
# Public functions
########################################################################

###########################
# executeAction() function
###########################


def executeAction(configPath, options, config):
    """
    Executes the amazons3 backup action.

    Args:
       configPath (String representing a path on disk): Path to configuration file on disk
       options (Options object): Program command-line options
       config (Config object): Program configuration
    Raises:
       ValueError: Under many generic error conditions
       IOError: If there are I/O problems reading or writing files
    """
    logger.debug("Executing amazons3 extended action.")
    if not isRunningAsRoot():
        logger.error("Error: the amazons3 extended action must be run as root.")
        raise ValueError("The amazons3 extended action must be run as root.")
    if sys.platform == "win32":
        logger.error("Error: the amazons3 extended action is not supported on Windows.")
        raise ValueError("The amazons3 extended action is not supported on Windows.")
    if config.options is None or config.stage is None:
        raise ValueError("Cedar Backup configuration is not properly filled in.")
    local = LocalConfig(xmlPath=configPath)
    stagingDirs = _findCorrectDailyDir(options, config, local)
    _applySizeLimits(options, config, local, stagingDirs)
    _writeToAmazonS3(config, local, stagingDirs)
    _writeStoreIndicator(config, stagingDirs)
    logger.info("Executed the amazons3 extended action successfully.")


########################################################################
# Private utility functions
########################################################################

#########################
# _findCorrectDailyDir()
#########################


def _findCorrectDailyDir(options, config, local):
    """
    Finds the correct daily staging directory to be written to Amazon S3.

    This is substantially similar to the same function in store.py.  The
    main difference is that it doesn't rely on store configuration at all.

    Args:
       options: Options object
       config: Config object
       local: Local config object

    Returns:
        Correct staging dir, as a dict mapping directory to date suffix
    Raises:
       IOError: If the staging directory cannot be found
    """
    oneDay = datetime.timedelta(days=1)
    today = datetime.date.today()
    yesterday = today - oneDay
    tomorrow = today + oneDay
    todayDate = today.strftime(DIR_TIME_FORMAT)
    yesterdayDate = yesterday.strftime(DIR_TIME_FORMAT)
    tomorrowDate = tomorrow.strftime(DIR_TIME_FORMAT)
    todayPath = pathJoin(config.stage.targetDir, todayDate)
    yesterdayPath = pathJoin(config.stage.targetDir, yesterdayDate)
    tomorrowPath = pathJoin(config.stage.targetDir, tomorrowDate)
    todayStageInd = pathJoin(todayPath, STAGE_INDICATOR)
    yesterdayStageInd = pathJoin(yesterdayPath, STAGE_INDICATOR)
    tomorrowStageInd = pathJoin(tomorrowPath, STAGE_INDICATOR)
    todayStoreInd = pathJoin(todayPath, STORE_INDICATOR)
    yesterdayStoreInd = pathJoin(yesterdayPath, STORE_INDICATOR)
    tomorrowStoreInd = pathJoin(tomorrowPath, STORE_INDICATOR)
    if options.full:
        if os.path.isdir(todayPath) and os.path.exists(todayStageInd):
            logger.info("Amazon S3 process will use current day's staging directory [%s]", todayPath)
            return {todayPath: todayDate}
        raise OSError("Unable to find staging directory to process (only tried today due to full option).")
    else:
        if os.path.isdir(todayPath) and os.path.exists(todayStageInd) and not os.path.exists(todayStoreInd):
            logger.info("Amazon S3 process will use current day's staging directory [%s]", todayPath)
            return {todayPath: todayDate}
        elif os.path.isdir(yesterdayPath) and os.path.exists(yesterdayStageInd) and not os.path.exists(yesterdayStoreInd):
            logger.info("Amazon S3 process will use previous day's staging directory [%s]", yesterdayPath)
            if local.amazons3.warnMidnite:
                logger.warning("Warning: Amazon S3 process crossed midnite boundary to find data.")
            return {yesterdayPath: yesterdayDate}
        elif os.path.isdir(tomorrowPath) and os.path.exists(tomorrowStageInd) and not os.path.exists(tomorrowStoreInd):
            logger.info("Amazon S3 process will use next day's staging directory [%s]", tomorrowPath)
            if local.amazons3.warnMidnite:
                logger.warning("Warning: Amazon S3 process crossed midnite boundary to find data.")
            return {tomorrowPath: tomorrowDate}
        raise OSError("Unable to find unused staging directory to process (tried today, yesterday, tomorrow).")


##############################
# _applySizeLimits() function
##############################


def _applySizeLimits(options, config, local, stagingDirs):
    """
    Apply size limits, throwing an exception if any limits are exceeded.

    Size limits are optional.  If a limit is set to None, it does not apply.
    The full size limit applies if the full option is set or if today is the
    start of the week.  The incremental size limit applies otherwise.  Limits
    are applied to the total size of all the relevant staging directories.

    Args:
       options: Options object
       config: Config object
       local: Local config object
       stagingDirs: Dictionary mapping directory path to date suffix

    Raises:
       ValueError: Under many generic error conditions
       ValueError: If a size limit has been exceeded
    """
    if options.full or isStartOfWeek(config.options.startingDay):
        logger.debug("Using Amazon S3 size limit for full backups.")
        limit = local.amazons3.fullBackupSizeLimit
    else:
        logger.debug("Using Amazon S3 size limit for incremental backups.")
        limit = local.amazons3.incrementalBackupSizeLimit
    if limit is None:
        logger.debug("No Amazon S3 size limit will be applied.")
    else:
        logger.debug("Amazon S3 size limit is: %s", limit)
        contents = BackupFileList()
        for stagingDir in stagingDirs:
            contents.addDirContents(stagingDir)
        total = contents.totalSize()
        logger.debug("Amazon S3 backup size is: %s", displayBytes(total))
        if total > limit:
            logger.error("Amazon S3 size limit exceeded: %s > %s", displayBytes(total), limit)
            raise ValueError("Amazon S3 size limit exceeded: %s > %s" % (displayBytes(total), limit))
        else:
            logger.info("Total size does not exceed Amazon S3 size limit, so backup can continue.")


##############################
# _writeToAmazonS3() function
##############################


def _writeToAmazonS3(config, local, stagingDirs):
    """
    Writes the indicated staging directories to an Amazon S3 bucket.

    Each of the staging directories listed in ``stagingDirs`` will be written to
    the configured Amazon S3 bucket from local configuration.  The directories
    will be placed into the image at the root by date, so staging directory
    ``/opt/stage/2005/02/10`` will be placed into the S3 bucket at ``/2005/02/10``.
    If an encrypt commmand is provided, the files will be encrypted first.

    Args:
       config: Config object
       local: Local config object
       stagingDirs: Dictionary mapping directory path to date suffix

    Raises:
       ValueError: Under many generic error conditions
       IOError: If there is a problem writing to Amazon S3
    """
    for stagingDir in list(stagingDirs.keys()):
        logger.debug("Storing stage directory to Amazon S3 [%s].", stagingDir)
        dateSuffix = stagingDirs[stagingDir]
        s3BucketUrl = "s3://%s/%s" % (local.amazons3.s3Bucket, dateSuffix)
        logger.debug("S3 bucket URL is [%s]", s3BucketUrl)
        _clearExistingBackup(config, s3BucketUrl)
        if local.amazons3.encryptCommand is None:
            logger.debug("Encryption is disabled; files will be uploaded in cleartext.")
            _uploadStagingDir(config, stagingDir, s3BucketUrl)
            _verifyUpload(config, stagingDir, s3BucketUrl)
        else:
            logger.debug("Encryption is enabled; files will be uploaded after being encrypted.")
            encryptedDir = tempfile.mkdtemp(dir=config.options.workingDir)
            changeOwnership(encryptedDir, config.options.backupUser, config.options.backupGroup)
            try:
                _encryptStagingDir(config, local, stagingDir, encryptedDir)
                _uploadStagingDir(config, encryptedDir, s3BucketUrl)
                _verifyUpload(config, encryptedDir, s3BucketUrl)
            finally:
                if os.path.exists(encryptedDir):
                    shutil.rmtree(encryptedDir)


##################################
# _writeStoreIndicator() function
##################################


def _writeStoreIndicator(config, stagingDirs):
    """
    Writes a store indicator file into staging directories.
    Args:
       config: Config object
       stagingDirs: Dictionary mapping directory path to date suffix
    """
    for stagingDir in list(stagingDirs.keys()):
        writeIndicatorFile(stagingDir, STORE_INDICATOR, config.options.backupUser, config.options.backupGroup)


##################################
# _clearExistingBackup() function
##################################


def _clearExistingBackup(config, s3BucketUrl):
    """
    Clear any existing backup files for an S3 bucket URL.
    Args:
       config: Config object
       s3BucketUrl: S3 bucket URL associated with the staging directory
    """
    suCommand = resolveCommand(SU_COMMAND)
    awsCommand = resolveCommand(AWS_COMMAND)
    actualCommand = "%s s3 rm --recursive %s/" % (awsCommand[0], s3BucketUrl)
    result = executeCommand(suCommand, [config.options.backupUser, "-c", actualCommand])[0]
    if result != 0:
        raise OSError("Error [%d] calling AWS CLI to clear existing backup for [%s]." % (result, s3BucketUrl))
    logger.debug("Completed clearing any existing backup in S3 for [%s]", s3BucketUrl)


###############################
# _uploadStagingDir() function
###############################


def _uploadStagingDir(config, stagingDir, s3BucketUrl):
    """
    Upload the contents of a staging directory out to the Amazon S3 cloud.
    Args:
       config: Config object
       stagingDir: Staging directory to upload
       s3BucketUrl: S3 bucket URL associated with the staging directory
    """
    # The version of awscli in Debian stretch (1.11.13-1) has a problem
    # uploading empty files, due to running with Python 3 rather than Python 2
    # as the upstream maintainers intended.  To work around this, I'm explicitly
    # excluding files like cback.stage, cback.collect, etc. which should be the
    # only empty files we ever try to copy.  See: https://github.com/aws/aws-cli/issues/2403
    suCommand = resolveCommand(SU_COMMAND)
    awsCommand = resolveCommand(AWS_COMMAND)
    actualCommand = '%s s3 cp --recursive --exclude "*cback.*" %s/ %s/' % (awsCommand[0], stagingDir, s3BucketUrl)
    result = executeCommand(suCommand, [config.options.backupUser, "-c", actualCommand])[0]
    if result != 0:
        raise OSError("Error [%d] calling AWS CLI to upload staging directory to [%s]." % (result, s3BucketUrl))
    logger.debug("Completed uploading staging dir [%s] to [%s]", stagingDir, s3BucketUrl)


###########################
# _verifyUpload() function
###########################


def _verifyUpload(config, stagingDir, s3BucketUrl):
    """
    Verify that a staging directory was properly uploaded to the Amazon S3 cloud.
    Args:
       config: Config object
       stagingDir: Staging directory to verify
       s3BucketUrl: S3 bucket URL associated with the staging directory
    """
    (bucket, prefix) = s3BucketUrl.replace("s3://", "").split("/", 1)
    suCommand = resolveCommand(SU_COMMAND)
    awsCommand = resolveCommand(AWS_COMMAND)
    query = "Contents[].{Key: Key, Size: Size}"
    actualCommand = "%s s3api list-objects --bucket %s --prefix %s --query '%s'" % (awsCommand[0], bucket, prefix, query)
    (result, data) = executeCommand(suCommand, [config.options.backupUser, "-c", actualCommand], returnOutput=True)
    if result != 0:
        raise OSError("Error [%d] calling AWS CLI verify upload to [%s]." % (result, s3BucketUrl))
    contents = {}
    for entry in json.loads("".join(data)):
        key = entry["Key"].replace(prefix, "")
        size = int(entry["Size"])
        contents[key] = size
    files = FilesystemList()
    files.excludeBasenamePatterns = [r"cback\..*"]  # because these are excluded from the upload
    files.addDirContents(stagingDir)
    for entry in files:
        if os.path.isfile(entry):
            key = entry.replace(stagingDir, "")
            size = int(os.stat(entry).st_size)
            if key not in contents:
                raise OSError("File was apparently not uploaded: [%s]" % entry)
            else:
                if size != contents[key]:
                    raise OSError("File size differs [%s], expected %s bytes but got %s bytes" % (entry, size, contents[key]))
    logger.debug("Completed verifying upload from [%s] to [%s].", stagingDir, s3BucketUrl)


################################
# _encryptStagingDir() function
################################


def _encryptStagingDir(config, local, stagingDir, encryptedDir):
    """
    Encrypt a staging directory, creating a new directory in the process.
    Args:
       config: Config object
       stagingDir: Staging directory to use as source
       encryptedDir: Target directory into which encrypted files should be written
    """
    suCommand = resolveCommand(SU_COMMAND)
    files = FilesystemList()
    files.addDirContents(stagingDir)
    for cleartext in files:
        if os.path.isfile(cleartext):
            encrypted = "%s%s" % (encryptedDir, cleartext.replace(stagingDir, ""))
            if int(os.stat(cleartext).st_size) == 0:
                with open(encrypted, "a") as f:
                    f.close()  # don't bother encrypting empty files
            else:
                actualCommand = local.amazons3.encryptCommand.replace("${input}", cleartext).replace("${output}", encrypted)
                subdir = os.path.dirname(encrypted)
                if not os.path.isdir(subdir):
                    os.makedirs(subdir)
                    changeOwnership(subdir, config.options.backupUser, config.options.backupGroup)
                result = executeCommand(suCommand, [config.options.backupUser, "-c", actualCommand])[0]
                if result != 0:
                    raise OSError("Error [%d] encrypting [%s]." % (result, cleartext))
    logger.debug("Completed encrypting staging directory [%s] into [%s]", stagingDir, encryptedDir)



# vim: set ft=python ts=4 sw=4 expandtab:
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2014-2016,2020 Kenneth J. Pronovici.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# Version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Copies of the GNU General Public License are available from
# the Free Software Foundation website, http://www.gnu.org/.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Cedar Backup, release 3
# Purpose  : Cedar Backup tool to synchronize an Amazon S3 bucket.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################################
# Notes
########################################################################

"""
Synchonizes a local directory with an Amazon S3 bucket.

No configuration is required; all necessary information is taken from the
command-line.  The only thing configuration would help with is the path
resolver interface, and it doesn't seem worth it to require configuration just
to get that.

:author: Kenneth J. Pronovici <pronovic@ieee.org>
"""

########################################################################
# Imported modules and constants
########################################################################

# Note: getopt is "soft deprecated" only and is safe to use; see: https://github.com/python/cpython/pull/105735

import getopt
import json
import logging
import os
import sys
import warnings
from functools import total_ordering
from pathlib import Path

import chardet

from CedarBackup3.cli import DEFAULT_LOGFILE, DEFAULT_MODE, DEFAULT_OWNERSHIP, setupLogging
from CedarBackup3.filesystem import FilesystemList
from CedarBackup3.release import AUTHOR, EMAIL, VERSION
from CedarBackup3.util import Diagnostics, encodePath, executeCommand, splitCommandLine

########################################################################
# Module-wide constants and variables
########################################################################

logger = logging.getLogger("CedarBackup3.log.tools.amazons3")

AWS_COMMAND = ["aws"]

SHORT_SWITCHES = "hVbql:o:m:OdsDvuw"
LONG_SWITCHES = [
    "help",
    "version",
    "verbose",
    "quiet",
    "logfile=",
    "owner=",
    "mode=",
    "output",
    "debug",
    "stack",
    "diagnostics",
    "verifyOnly",
    "uploadOnly",
    "ignoreWarnings",
]


#######################################################################
# Options class
#######################################################################


@total_ordering
class Options:
    ######################
    # Class documentation
    ######################

    """
    Class representing command-line options for the cback3-amazons3-sync script.

    The ``Options`` class is a Python object representation of the command-line
    options of the cback3-amazons3-sync script.

    The object representation is two-way: a command line string or a list of
    command line arguments can be used to create an ``Options`` object, and then
    changes to the object can be propogated back to a list of command-line
    arguments or to a command-line string.  An ``Options`` object can even be
    created from scratch programmatically (if you have a need for that).

    There are two main levels of validation in the ``Options`` class.  The first
    is field-level validation.  Field-level validation comes into play when a
    given field in an object is assigned to or updated.  We use Python's
    ``property`` functionality to enforce specific validations on field values,
    and in some places we even use customized list classes to enforce
    validations on list members.  You should expect to catch a ``ValueError``
    exception when making assignments to fields if you are programmatically
    filling an object.

    The second level of validation is post-completion validation.  Certain
    validations don't make sense until an object representation of options is
    fully "complete".  We don't want these validations to apply all of the time,
    because it would make building up a valid object from scratch a real pain.
    For instance, we might have to do things in the right order to keep from
    throwing exceptions, etc.

    All of these post-completion validations are encapsulated in the
    :any:`Options.validate` method.  This method can be called at any time by a
    client, and will always be called immediately after creating a ``Options``
    object from a command line and before exporting a ``Options`` object back to
    a command line.  This way, we get acceptable ease-of-use but we also don't
    accept or emit invalid command lines.

    *Note:* Lists within this class are "unordered" for equality comparisons.

    """

    ##############
    # Constructor
    ##############

    def __init__(self, argumentList=None, argumentString=None, validate=True):
        """
        Initializes an options object.

        If you initialize the object without passing either ``argumentList`` or
        ``argumentString``, the object will be empty and will be invalid until it
        is filled in properly.

        No reference to the original arguments is saved off by this class.  Once
        the data has been parsed (successfully or not) this original information
        is discarded.

        The argument list is assumed to be a list of arguments, not including the
        name of the command, something like ``sys.argv[1:]``.  If you pass
        ``sys.argv`` instead, things are not going to work.

        The argument string will be parsed into an argument list by the
        :any:`util.splitCommandLine` function (see the documentation for that
        function for some important notes about its limitations).  There is an
        assumption that the resulting list will be equivalent to ``sys.argv[1:]``,
        just like ``argumentList``.

        Unless the ``validate`` argument is ``False``, the :any:`Options.validate`
        method will be called (with its default arguments) after successfully
        parsing any passed-in command line.  This validation ensures that
        appropriate actions, etc. have been specified.  Keep in mind that even if
        ``validate`` is ``False``, it might not be possible to parse the passed-in
        command line, so an exception might still be raised.

        *Note:* The command line format is specified by the ``_usage`` function.
        Call ``_usage`` to see a usage statement for the cback3-amazons3-sync script.

        *Note:* It is strongly suggested that the ``validate`` option always be set
        to ``True`` (the default) unless there is a specific need to read in
        invalid command line arguments.

        Args:
           argumentList (List of arguments, i.e. ``sys.argv``): Command line for a program
           argumentString (String, i.e. "cback3-amazons3-sync --verbose stage store"): Command line for a program
           validate (Boolean true/false): Validate the command line after parsing it
        Raises:
           getopt.GetoptError: If the command-line arguments could not be parsed
           ValueError: If the command-line arguments are invalid
        """
        self._help = False
        self._version = False
        self._verbose = False
        self._quiet = False
        self._logfile = None
        self._owner = None
        self._mode = None
        self._output = False
        self._debug = False
        self._stacktrace = False
        self._diagnostics = False
        self._verifyOnly = False
        self._uploadOnly = False
        self._ignoreWarnings = False
        self._sourceDir = None
        self._s3BucketUrl = None
        if argumentList is not None and argumentString is not None:
            raise ValueError("Use either argumentList or argumentString, but not both.")
        if argumentString is not None:
            argumentList = splitCommandLine(argumentString)
        if argumentList is not None:
            self._parseArgumentList(argumentList)
            if validate:
                self.validate()

    #########################
    # String representations
    #########################

    def __repr__(self):
        """
        Official string representation for class instance.
        """
        return self.buildArgumentString(validate=False)

    def __str__(self):
        """
        Informal string representation for class instance.
        """
        return self.__repr__()

    #############################
    # Standard comparison method
    #############################

    def __eq__(self, other):
        """Equals operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        """Less-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) < 0

    def __gt__(self, other):
        """Greater-than operator, iplemented in terms of original Python 2 compare operator."""
        return self.__cmp__(other) > 0

    def __cmp__(self, other):
        """
        Original Python 2 comparison operator.
        Lists within this class are "unordered" for equality comparisons.
        Args:
           other: Other object to compare to
        Returns:
            -1/0/1 depending on whether self is ``<``, ``=`` or ``>`` other
        """
        if other is None:
            return 1
        if self.help != other.help:
            if self.help < other.help:
                return -1
            else:
                return 1
        if self.version != other.version:
            if self.version < other.version:
                return -1
            else:
                return 1
        if self.verbose != other.verbose:
            if self.verbose < other.verbose:
                return -1
            else:
                return 1
        if self.quiet != other.quiet:
            if self.quiet < other.quiet:
                return -1
            else:
                return 1
        if self.logfile != other.logfile:
            if str(self.logfile or "") < str(other.logfile or ""):
                return -1
            else:
                return 1
        if self.owner != other.owner:
            if str(self.owner or "") < str(other.owner or ""):
                return -1
            else:
                return 1
        if self.mode != other.mode:
            if int(self.mode or 0) < int(other.mode or 0):
                return -1
            else:
                return 1
        if self.output != other.output:
            if self.output < other.output:
                return -1
            else:
                return 1
        if self.debug != other.debug:
            if self.debug < other.debug:
                return -1
            else:
                return 1
        if self.stacktrace != other.stacktrace:
            if self.stacktrace < other.stacktrace:
                return -1
            else:
                return 1
        if self.diagnostics != other.diagnostics:
            if self.diagnostics < other.diagnostics:
                return -1
            else:
                return 1
        if self.verifyOnly != other.verifyOnly:
            if self.verifyOnly < other.verifyOnly:
                return -1
            else:
                return 1
        if self.uploadOnly != other.uploadOnly:
            if self.uploadOnly < other.uploadOnly:
                return -1
            else:
                return 1
        if self.ignoreWarnings != other.ignoreWarnings:
            if self.ignoreWarnings < other.ignoreWarnings:
                return -1
            else:
                return 1
        if self.sourceDir != other.sourceDir:
            if str(self.sourceDir or "") < str(other.sourceDir or ""):
                return -1
            else:
                return 1
        if self.s3BucketUrl != other.s3BucketUrl:
            if str(self.s3BucketUrl or "") < str(other.s3BucketUrl or ""):
                return -1
            else:
                return 1
        return 0

    #############
    # Properties
    #############

    def _setHelp(self, value):
        """
        Property target used to set the help flag.
        No validations, but we normalize the value to ``True`` or ``False``.
        """
        if value:
            self._help = True
        else:
            self._help = False

    def _getHelp(self):
        """
        Property target used to get the help flag.
        """
        return self._help

    def _setVersion(self, value):
        """
        Property target used to set the version flag.
        No validations, but we normalize the value to ``True`` or ``False``.
        """
        if value:
            self._version = True
        else:
            self._version = False

    def _getVersion(self):
        """
        Property target used to get the version flag.
        """
        return self._version

    def _setVerbose(self, value):
        """
        Property target used to set the verbose flag.
        No validations, but we normalize the value to ``True`` or ``False``.
        """
        if value:
            self._verbose = True
        else:
            self._verbose = False

    def _getVerbose(self):
        """
        Property target used to get the verbose flag.
        """
        return self._verbose

    def _setQuiet(self, value):
        """
        Property target used to set the quiet flag.
        No validations, but we normalize the value to ``True`` or ``False``.
        """
        if value:
            self._quiet = True
        else:
            self._quiet = False

    def _getQuiet(self):
        """
        Property target used to get the quiet flag.
        """
        return self._quiet

    def _setLogfile(self, value):
        """
        Property target used to set the logfile parameter.
        Raises:
           ValueError: If the value cannot be encoded properly
        """
        if value is not None:
            if len(value) < 1:
                raise ValueError("The logfile parameter must be a non-empty string.")
        self._logfile = encodePath(value)

    def _getLogfile(self):
        """
        Property target used to get the logfile parameter.
        """
        return self._logfile

    def _setOwner(self, value):
        """
        Property target used to set the owner parameter.
        If not ``None``, the owner must be a ``(user,group)`` tuple or list.
        Strings (and inherited children of strings) are explicitly disallowed.
        The value will be normalized to a tuple.
        Raises:
           ValueError: If the value is not valid
        """
        if value is None:
            self._owner = None
        else:
            if isinstance(value, str):
                raise ValueError("Must specify user and group tuple for owner parameter.")
            if len(value) != 2:
                raise ValueError("Must specify user and group tuple for owner parameter.")
            if len(value[0]) < 1 or len(value[1]) < 1:
                raise ValueError("User and group tuple values must be non-empty strings.")
            self._owner = (value[0], value[1])

    def _getOwner(self):
        """
        Property target used to get the owner parameter.
        The parameter is a tuple of ``(user, group)``.
        """
        return self._owner

    def _setMode(self, value):
        """
        Property target used to set the mode parameter.
        """
        if value is None:
            self._mode = None
        else:
            try:
                if isinstance(value, str):
                    value = int(value, 8)
                else:
                    value = int(value)
            except TypeError:
                raise ValueError("Mode must be an octal integer >= 0, i.e. 644.")
            if value < 0:
                raise ValueError("Mode must be an octal integer >= 0. i.e. 644.")
            self._mode = value

    def _getMode(self):
        """
        Property target used to get the mode parameter.
        """
        return self._mode

    def _setOutput(self, value):
        """
        Property target used to set the output flag.
        No validations, but we normalize the value to ``True`` or ``False``.
        """
        if value:
            self._output = True
        else:
            self._output = False

    def _getOutput(self):
        """
        Property target used to get the output flag.
        """
        return self._output

    def _setDebug(self, value):
        """
        Property target used to set the debug flag.
        No validations, but we normalize the value to ``True`` or ``False``.
        """
        if value:
            self._debug = True
        else:
            self._debug = False

    def _getDebug(self):
        """
        Property target used to get the debug flag.
        """
        return self._debug

    def _setStacktrace(self, value):
        """
        Property target used to set the stacktrace flag.
        No validations, but we normalize the value to ``True`` or ``False``.
        """
        if value:
            self._stacktrace = True
        else:
            self._stacktrace = False

    def _getStacktrace(self):
        """
        Property target used to get the stacktrace flag.
        """
        return self._stacktrace

    def _setDiagnostics(self, value):
        """
        Property target used to set the diagnostics flag.
        No validations, but we normalize the value to ``True`` or ``False``.
        """
        if value:
            self._diagnostics = True
        else:
            self._diagnostics = False

    def _getDiagnostics(self):
        """
        Property target used to get the diagnostics flag.
        """
        return self._diagnostics

    def _setVerifyOnly(self, value):
        """
        Property target used to set the verifyOnly flag.
        No validations, but we normalize the value to ``True`` or ``False``.
        """
        if value:
            self._verifyOnly = True
        else:
            self._verifyOnly = False

    def _getVerifyOnly(self):
        """
        Property target used to get the verifyOnly flag.
        """
        return self._verifyOnly

    def _setUploadOnly(self, value):
        """
        Property target used to set the uploadOnly flag.
        No validations, but we normalize the value to ``True`` or ``False``.
        """
        if value:
            self._uploadOnly = True
        else:
            self._uploadOnly = False

    def _getUploadOnly(self):
        """
        Property target used to get the uploadOnly flag.
        """
        return self._uploadOnly

    def _setIgnoreWarnings(self, value):
        """
        Property target used to set the ignoreWarnings flag.
        No validations, but we normalize the value to ``True`` or ``False``.
        """
        if value:
            self._ignoreWarnings = True
        else:
            self._ignoreWarnings = False

    def _getIgnoreWarnings(self):
        """
        Property target used to get the ignoreWarnings flag.
        """
        return self._ignoreWarnings

    def _setSourceDir(self, value):
        """
        Property target used to set the sourceDir parameter.
        """
        if value is not None:
            if len(value) < 1:
                raise ValueError("The sourceDir parameter must be a non-empty string.")
        self._sourceDir = value

    def _getSourceDir(self):
        """
        Property target used to get the sourceDir parameter.
        """
        return self._sourceDir

    def _setS3BucketUrl(self, value):
        """
        Property target used to set the s3BucketUrl parameter.
        """
        if value is not None:
            if len(value) < 1:
                raise ValueError("The s3BucketUrl parameter must be a non-empty string.")
        self._s3BucketUrl = value

    def _getS3BucketUrl(self):
        """
        Property target used to get the s3BucketUrl parameter.
        """
        return self._s3BucketUrl

    help = property(_getHelp, _setHelp, None, "Command-line help (``-h,--help``) flag.")
    version = property(_getVersion, _setVersion, None, "Command-line version (``-V,--version``) flag.")
    verbose = property(_getVerbose, _setVerbose, None, "Command-line verbose (``-b,--verbose``) flag.")
    quiet = property(_getQuiet, _setQuiet, None, "Command-line quiet (``-q,--quiet``) flag.")
    logfile = property(_getLogfile, _setLogfile, None, "Command-line logfile (``-l,--logfile``) parameter.")
    owner = property(_getOwner, _setOwner, None, "Command-line owner (``-o,--owner``) parameter, as tuple ``(user,group)``.")
    mode = property(_getMode, _setMode, None, "Command-line mode (``-m,--mode``) parameter.")
    output = property(_getOutput, _setOutput, None, "Command-line output (``-O,--output``) flag.")
    debug = property(_getDebug, _setDebug, None, "Command-line debug (``-d,--debug``) flag.")
    stacktrace = property(_getStacktrace, _setStacktrace, None, "Command-line stacktrace (``-s,--stack``) flag.")
    diagnostics = property(_getDiagnostics, _setDiagnostics, None, "Command-line diagnostics (``-D,--diagnostics``) flag.")
    verifyOnly = property(_getVerifyOnly, _setVerifyOnly, None, "Command-line verifyOnly (``-v,--verifyOnly``) flag.")
    uploadOnly = property(_getUploadOnly, _setUploadOnly, None, "Command-line uploadOnly (``-u,--uploadOnly``) flag.")
    ignoreWarnings = property(
        _getIgnoreWarnings, _setIgnoreWarnings, None, "Command-line ignoreWarnings (``-w,--ignoreWarnings``) flag"
    )
    sourceDir = property(_getSourceDir, _setSourceDir, None, "Command-line sourceDir, source of sync.")
    s3BucketUrl = property(_getS3BucketUrl, _setS3BucketUrl, None, "Command-line s3BucketUrl, target of sync.")

    ##################
    # Utility methods
    ##################

    def validate(self):
        """
        Validates command-line options represented by the object.

        Unless ``--help`` or ``--version`` are supplied, at least one action must
        be specified.  Other validations (as for allowed values for particular
        options) will be taken care of at assignment time by the properties
        functionality.

        *Note:* The command line format is specified by the ``_usage`` function.
        Call ``_usage`` to see a usage statement for the cback3-amazons3-sync script.

        Raises:
           ValueError: If one of the validations fails
        """
        if not self.help and not self.version and not self.diagnostics:
            if self.sourceDir is None or self.s3BucketUrl is None:
                raise ValueError("Source directory and S3 bucket URL are both required.")

    def buildArgumentList(self, validate=True):
        """
        Extracts options into a list of command line arguments.

        The original order of the various arguments (if, indeed, the object was
        initialized with a command-line) is not preserved in this generated
        argument list.   Besides that, the argument list is normalized to use the
        long option names (i.e. --version rather than -V).  The resulting list
        will be suitable for passing back to the constructor in the
        ``argumentList`` parameter.  Unlike :any:`buildArgumentString`, string
        arguments are not quoted here, because there is no need for it.

        Unless the ``validate`` parameter is ``False``, the :any:`Options.validate`
        method will be called (with its default arguments) against the
        options before extracting the command line.  If the options are not valid,
        then an argument list will not be extracted.

        *Note:* It is strongly suggested that the ``validate`` option always be set
        to ``True`` (the default) unless there is a specific need to extract an
        invalid command line.

        Args:
           validate (Boolean true/false): Validate the options before extracting the command line
        Returns:
            List representation of command-line arguments
        Raises:
           ValueError: If options within the object are invalid
        """
        if validate:
            self.validate()
        argumentList = []
        if self._help:
            argumentList.append("--help")
        if self.version:
            argumentList.append("--version")
        if self.verbose:
            argumentList.append("--verbose")
        if self.quiet:
            argumentList.append("--quiet")
        if self.logfile is not None:
            argumentList.append("--logfile")
            argumentList.append(self.logfile)
        if self.owner is not None:
            argumentList.append("--owner")
            argumentList.append("%s:%s" % (self.owner[0], self.owner[1]))
        if self.mode is not None:
            argumentList.append("--mode")
            argumentList.append("%o" % self.mode)
        if self.output:
            argumentList.append("--output")
        if self.debug:
            argumentList.append("--debug")
        if self.stacktrace:
            argumentList.append("--stack")
        if self.diagnostics:
            argumentList.append("--diagnostics")
        if self.verifyOnly:
            argumentList.append("--verifyOnly")
        if self.uploadOnly:
            argumentList.append("--uploadOnly")
        if self.ignoreWarnings:
            argumentList.append("--ignoreWarnings")
        if self.sourceDir is not None:
            argumentList.append(self.sourceDir)
        if self.s3BucketUrl is not None:
            argumentList.append(self.s3BucketUrl)
        return argumentList

    def buildArgumentString(self, validate=True):
        """
        Extracts options into a string of command-line arguments.

        The original order of the various arguments (if, indeed, the object was
        initialized with a command-line) is not preserved in this generated
        argument string.   Besides that, the argument string is normalized to use
        the long option names (i.e. --version rather than -V) and to quote all
        string arguments with double quotes (``"``).  The resulting string will be
        suitable for passing back to the constructor in the ``argumentString``
        parameter.

        Unless the ``validate`` parameter is ``False``, the :any:`Options.validate`
        method will be called (with its default arguments) against the options
        before extracting the command line.  If the options are not valid, then
        an argument string will not be extracted.

        *Note:* It is strongly suggested that the ``validate`` option always be set
        to ``True`` (the default) unless there is a specific need to extract an
        invalid command line.

        Args:
           validate (Boolean true/false): Validate the options before extracting the command line
        Returns:
            String representation of command-line arguments
        Raises:
           ValueError: If options within the object are invalid
        """
        if validate:
            self.validate()
        argumentString = ""
        if self._help:
            argumentString += "--help "
        if self.version:
            argumentString += "--version "
        if self.verbose:
            argumentString += "--verbose "
        if self.quiet:
            argumentString += "--quiet "
        if self.logfile is not None:
            argumentString += '--logfile "%s" ' % self.logfile
        if self.owner is not None:
            argumentString += '--owner "%s:%s" ' % (self.owner[0], self.owner[1])
        if self.mode is not None:
            argumentString += "--mode %o " % self.mode
        if self.output:
            argumentString += "--output "
        if self.debug:
            argumentString += "--debug "
        if self.stacktrace:
            argumentString += "--stack "
        if self.diagnostics:
            argumentString += "--diagnostics "
        if self.verifyOnly:
            argumentString += "--verifyOnly "
        if self.uploadOnly:
            argumentString += "--uploadOnly "
        if self.ignoreWarnings:
            argumentString += "--ignoreWarnings "
        if self.sourceDir is not None:
            argumentString += '"%s" ' % self.sourceDir
        if self.s3BucketUrl is not None:
            argumentString += '"%s" ' % self.s3BucketUrl
        return argumentString

    def _parseArgumentList(self, argumentList):
        """
        Internal method to parse a list of command-line arguments.

        Most of the validation we do here has to do with whether the arguments
        can be parsed and whether any values which exist are valid.  We don't do
        any validation as to whether required elements exist or whether elements
        exist in the proper combination (instead, that's the job of the
        :any:`validate` method).

        For any of the options which supply parameters, if the option is
        duplicated with long and short switches (i.e. ``-l`` and a ``--logfile``)
        then the long switch is used.  If the same option is duplicated with the
        same switch (long or short), then the last entry on the command line is
        used.

        Args:
           argumentList (List of arguments to a command, i.e. ``sys.argv[1:]``): List of arguments to a command
        Raises:
           ValueError: If the argument list cannot be successfully parsed
        """
        switches = {}
        opts, remaining = getopt.getopt(argumentList, SHORT_SWITCHES, LONG_SWITCHES)
        for o, a in opts:  # push the switches into a hash
            switches[o] = a
        if "-h" in switches or "--help" in switches:
            self.help = True
        if "-V" in switches or "--version" in switches:
            self.version = True
        if "-b" in switches or "--verbose" in switches:
            self.verbose = True
        if "-q" in switches or "--quiet" in switches:
            self.quiet = True
        if "-l" in switches:
            self.logfile = switches["-l"]
        if "--logfile" in switches:
            self.logfile = switches["--logfile"]
        if "-o" in switches:
            self.owner = switches["-o"].split(":", 1)
        if "--owner" in switches:
            self.owner = switches["--owner"].split(":", 1)
        if "-m" in switches:
            self.mode = switches["-m"]
        if "--mode" in switches:
            self.mode = switches["--mode"]
        if "-O" in switches or "--output" in switches:
            self.output = True
        if "-d" in switches or "--debug" in switches:
            self.debug = True
        if "-s" in switches or "--stack" in switches:
            self.stacktrace = True
        if "-D" in switches or "--diagnostics" in switches:
            self.diagnostics = True
        if "-v" in switches or "--verifyOnly" in switches:
            self.verifyOnly = True
        if "-u" in switches or "--uploadOnly" in switches:
            self.uploadOnly = True
        if "-w" in switches or "--ignoreWarnings" in switches:
            self.ignoreWarnings = True
        try:
            (self.sourceDir, self.s3BucketUrl) = remaining
        except ValueError:
            pass


#######################################################################
# Public functions
#######################################################################

#################
# cli() function
#################


def cli():
    """
    Implements the command-line interface for the ``cback3-amazons3-sync`` script.

    Essentially, this is the "main routine" for the cback3-amazons3-sync script.  It does
    all of the argument processing for the script, and then also implements the
    tool functionality.

    This function looks pretty similiar to ``CedarBackup3.cli.cli()``.  It's not
    easy to refactor this code to make it reusable and also readable, so I've
    decided to just live with the duplication.

    A different error code is returned for each type of failure:

       - ``1``: The Python interpreter version is not supported
       - ``2``: Error processing command-line arguments
       - ``3``: Error configuring logging
       - ``5``: Backup was interrupted with a CTRL-C or similar
       - ``6``: Error executing other parts of the script

    *Note:* This script uses print rather than logging to the INFO level, because
    it is interactive.  Underlying Cedar Backup functionality uses the logging
    mechanism exclusively.

    Returns:
        Error code as described above
    """
    try:
        if list(map(int, [sys.version_info[0], sys.version_info[1]])) < [3, 8]:
            sys.stderr.write("Python 3 version 3.8 or greater required.\n")
            return 1
    except:
        # sys.version_info isn't available before 2.0
        sys.stderr.write("Python 3 version 3.8 or greater required.\n")
        return 1

    try:
        options = Options(argumentList=sys.argv[1:])
    except Exception as e:
        _usage()
        sys.stderr.write(" *** Error: %s\n" % e)
        return 2

    if options.help:
        _usage()
        return 0
    if options.version:
        _version()
        return 0
    if options.diagnostics:
        _diagnostics()
        return 0

    if options.stacktrace:
        logfile = setupLogging(options)
    else:
        try:
            logfile = setupLogging(options)
        except Exception as e:
            sys.stderr.write("Error setting up logging: %s\n" % e)
            return 3

    logger.info("Cedar Backup Amazon S3 sync run started.")
    logger.info("Options were [%s]", options)
    logger.info("Logfile is [%s]", logfile)
    Diagnostics().logDiagnostics(method=logger.info)

    if options.stacktrace:
        _executeAction(options)
    else:
        try:
            _executeAction(options)
        except KeyboardInterrupt:
            logger.error("Backup interrupted.")
            logger.info("Cedar Backup Amazon S3 sync run completed with status 5.")
            return 5
        except Exception as e:
            logger.error("Error executing backup: %s", e)
            logger.info("Cedar Backup Amazon S3 sync run completed with status 6.")
            return 6

    logger.info("Cedar Backup Amazon S3 sync run completed with status 0.")
    return 0


#######################################################################
# Utility functions
#######################################################################

####################
# _usage() function
####################


def _usage(fd=sys.stderr):
    """
    Prints usage information for the cback3-amazons3-sync script.
    Args:
       fd: File descriptor used to print information
    *Note:* The ``fd`` is used rather than ``print`` to facilitate unit testing.
    """
    fd.write("\n")
    fd.write(" Usage: cback3-amazons3-sync [switches] sourceDir s3bucketUrl\n")
    fd.write("\n")
    fd.write(" Cedar Backup Amazon S3 sync tool.\n")
    fd.write("\n")
    fd.write(" This Cedar Backup utility synchronizes a local directory to an Amazon S3\n")
    fd.write(" bucket.  After the sync is complete, a validation step is taken.  An\n")
    fd.write(" error is reported if the contents of the bucket do not match the\n")
    fd.write(" source directory, or if the indicated size for any file differs.\n")
    fd.write(" This tool is a wrapper over the AWS CLI command-line tool.\n")
    fd.write("\n")
    fd.write(" The following arguments are required:\n")
    fd.write("\n")
    fd.write("   sourceDir            The local source directory on disk (must exist)\n")
    fd.write("   s3BucketUrl          The URL to the target Amazon S3 bucket\n")
    fd.write("\n")
    fd.write(" The following switches are accepted:\n")
    fd.write("\n")
    fd.write("   -h, --help           Display this usage/help listing\n")
    fd.write("   -V, --version        Display version information\n")
    fd.write("   -b, --verbose        Print verbose output as well as logging to disk\n")
    fd.write("   -q, --quiet          Run quietly (display no output to the screen)\n")
    fd.write("   -l, --logfile        Path to logfile (default: %s)\n" % DEFAULT_LOGFILE)
    fd.write(
        "   -o, --owner          Logfile ownership, user:group (default: %s:%s)\n" % (DEFAULT_OWNERSHIP[0], DEFAULT_OWNERSHIP[1])
    )
    fd.write("   -m, --mode           Octal logfile permissions mode (default: %o)\n" % DEFAULT_MODE)
    fd.write("   -O, --output         Record some sub-command (i.e. aws) output to the log\n")
    fd.write("   -d, --debug          Write debugging information to the log (implies --output)\n")
    fd.write(
        "   -s, --stack          Dump Python stack trace instead of swallowing exceptions\n"
    )  # exactly 80 characters in width!
    fd.write("   -D, --diagnostics    Print runtime diagnostics to the screen and exit\n")
    fd.write("   -v, --verifyOnly     Only verify the S3 bucket contents, do not make changes\n")
    fd.write("   -u, --uploadOnly     Only upload new data, do not remove files in the S3 bucket\n")
    fd.write("   -w, --ignoreWarnings Ignore warnings about problematic filename encodings\n")
    fd.write("\n")
    fd.write(" Typical usage would be something like:\n")
    fd.write("\n")
    fd.write("   cback3-amazons3-sync /home/myuser s3://example.com-backup/myuser\n")
    fd.write("\n")
    fd.write(" This will sync the contents of /home/myuser into the indicated bucket.\n")
    fd.write("\n")


######################
# _version() function
######################


def _version(fd=sys.stdout):
    """
    Prints version information for the cback3-amazons3-sync script.
    Args:
       fd: File descriptor used to print information
    *Note:* The ``fd`` is used rather than ``print`` to facilitate unit testing.
    """
    fd.write("\n")
    fd.write(" Cedar Backup Amazon S3 sync tool.\n")
    fd.write(" Included with Cedar Backup version %s.\n" % VERSION)
    fd.write("\n")
    fd.write(" Copyright (c) %s <%s>.\n" % (AUTHOR, EMAIL))
    fd.write(" See NOTICE for a list of included code and other contributors.\n")
    fd.write(" This is free software; there is NO warranty.  See the\n")
    fd.write(" GNU General Public License version 2 for copying conditions.\n")
    fd.write("\n")
    fd.write(" Use the --help option for usage information.\n")
    fd.write("\n")


##########################
# _diagnostics() function
##########################


def _diagnostics(fd=sys.stdout):
    """
    Prints runtime diagnostics information.
    Args:
       fd: File descriptor used to print information
    *Note:* The ``fd`` is used rather than ``print`` to facilitate unit testing.
    """
    fd.write("\n")
    fd.write("Diagnostics:\n")
    fd.write("\n")
    Diagnostics().printDiagnostics(fd=fd, prefix="   ")
    fd.write("\n")


############################
# _executeAction() function
############################


def _executeAction(options):
    """
    Implements the guts of the cback3-amazons3-sync tool.

    Args:
       options (Options object): Program command-line options
    Raises:
       Exception: Under many generic error conditions
    """
    sourceFiles = _buildSourceFiles(options.sourceDir)
    if not options.ignoreWarnings:
        _checkSourceFiles(options.sourceDir, sourceFiles)
    if not options.verifyOnly:
        _synchronizeBucket(options.sourceDir, options.s3BucketUrl, options.uploadOnly)
    _verifyBucketContents(options.sourceDir, sourceFiles, options.s3BucketUrl)


################################
# _buildSourceFiles() function
################################


def _buildSourceFiles(sourceDir):
    """
    Build a list of files in a source directory
    Args:
       sourceDir: Local source directory
    Returns:
        FilesystemList with contents of source directory
    """
    if not os.path.isdir(sourceDir):
        raise ValueError("Source directory does not exist on disk.")
    sourceFiles = FilesystemList()
    sourceFiles.addDirContents(sourceDir)
    return sourceFiles


###############################
# _checkSourceFiles() function
###############################


def _checkSourceFiles(sourceDir, sourceFiles):  # noqa: ARG001
    """
    Check source files, trying to guess which ones will have encoding problems.
    Args:
       sourceDir: Local source directory
       sourceDir: Local source directory
    @raises ValueError: If a problem file is found
    @see U{http://opensourcehacker.com/2011/09/16/fix-linux-filename-encodings-with-python/}
    @see U{http://serverfault.com/questions/82821/how-to-tell-the-language-encoding-of-a-filename-on-linux}
    @see U{http://randysofia.com/2014/06/06/aws-cli-and-your-locale/}
    """
    with warnings.catch_warnings():
        encoding = Diagnostics().encoding

        # Note: this was difficult to fully test.  As of the original Python 2
        # implementation, I had a bunch of files on disk that had inconsistent
        # encodings, so I was able to prove that the check warned about these
        # files initially, and then didn't warn after I fixed them.  I didn't
        # save off those files for a unit test (ugh) so by the time of the Python
        # 3 conversion -- which is subtly different because of the different way
        # Python 3 handles unicode strings -- I had to contrive some tests.  I
        # think the tests I wrote are consistent with the earlier problems, and I
        # do get the same result for those tests in both CedarBackup 2 and Cedar
        # Backup 3.  However, I can't be certain the implementation is
        # equivalent.  If someone runs into a situation that this code doesn't
        # handle, you may need to revisit the implementation.

        failed = False
        for entry in sourceFiles:
            path = bytes(Path(entry))
            result = chardet.detect(path)
            source = path.decode(result["encoding"])
            try:
                target = path.decode(encoding)
                if source != target:
                    logger.error("Inconsistent encoding for [%s]: got %s, but need %s", path, result["encoding"], encoding)
                    failed = True
            except Exception:
                logger.error("Inconsistent encoding for [%s]: got %s, but need %s", path, result["encoding"], encoding)
                failed = True

        if not failed:
            logger.info("Completed checking source filename encoding (no problems found).")
        else:
            logger.error("Some filenames have inconsistent encodings and will likely cause sync problems.")
            logger.error("You may be able to fix this by setting a more sensible locale in your environment.")
            logger.error("Aternately, you can rename the problem files to be valid in the indicated locale.")
            logger.error("To ignore this warning and proceed anyway, use --ignoreWarnings")
            raise ValueError("Some filenames have inconsistent encodings and will likely cause sync problems.")


################################
# _synchronizeBucket() function
################################


def _synchronizeBucket(sourceDir, s3BucketUrl, uploadOnly):
    """
    Synchronize a local directory to an Amazon S3 bucket.
    Args:
       sourceDir: Local source directory
       s3BucketUrl: Target S3 bucket URL
    """
    # Since at least early 2015, 'aws s3 sync' is always recursive and the
    # --recursive option is useless.  They eventually removed it and now using
    # it causes an error.  See: https://github.com/aws/aws-cli/issues/1170
    logger.info("Synchronizing local source directory up to Amazon S3.")
    args = ["s3", "sync", sourceDir, s3BucketUrl]
    if uploadOnly:
        logger.info("Sync process will only upload new data, never removing files in S3")
    else:
        logger.info("This will be a full sync process, removing any S3 files that do not exist in the source")
        args += ["--delete"]
    result = executeCommand(AWS_COMMAND, args, returnOutput=False)[0]
    if result != 0:
        raise OSError("Error [%d] calling AWS CLI synchronize bucket." % result)


###################################
# _verifyBucketContents() function
###################################


def _verifyBucketContents(sourceDir, sourceFiles, s3BucketUrl):
    """
    Verify that a source directory is equivalent to an Amazon S3 bucket.
    Args:
       sourceDir: Local source directory
       sourceFiles: Filesystem list containing contents of source directory
       s3BucketUrl: Target S3 bucket URL
    """
    # As of this writing, the documentation for the S3 API that we're using
    # below says that up to 1000 elements at a time are returned, and that we
    # have to manually handle pagination by looking for the IsTruncated element.
    # However, in practice, this is not true.  I have been testing with
    # "aws-cli/1.4.4 Python/2.7.3 Linux/3.2.0-4-686-pae", installed through PIP.
    # No matter how many items exist in my bucket and prefix, I get back a
    # single JSON result.  I've tested with buckets containing nearly 6000
    # elements.
    #
    # If I turn on debugging, it's clear that underneath, something in the API
    # is executing multiple list-object requests against AWS, and stiching
    # results together to give me back the final JSON result.  The debug output
    # clearly incldues multiple requests, and each XML response (except for the
    # final one) contains <IsTruncated>true</IsTruncated>.
    #
    # This feature is not mentioned in the offical changelog for any of the
    # releases going back to 1.0.0.  It appears to happen in the botocore
    # library, but I'll admit I can't actually find the code that implements it.
    # For now, all I can do is rely on this behavior and hope that the
    # documentation is out-of-date.  I'm not going to write code that tries to
    # parse out IsTruncated if I can't actually test that code.

    (bucket, prefix) = s3BucketUrl.replace("s3://", "").split("/", 1)

    query = "Contents[].{Key: Key, Size: Size}"
    args = ["s3api", "list-objects", "--bucket", bucket, "--prefix", prefix, "--query", query]
    (result, data) = executeCommand(AWS_COMMAND, args, returnOutput=True)
    if result != 0:
        raise OSError("Error [%d] calling AWS CLI verify bucket contents." % result)

    contents = {}
    for entry in json.loads("".join(data)):
        key = entry["Key"].replace(prefix, "")
        size = int(entry["Size"])
        contents[key] = size

    failed = False
    for entry in sourceFiles:
        if os.path.isfile(entry):
            key = entry.replace(sourceDir, "")
            size = int(os.stat(entry).st_size)
            if key not in contents:
                logger.error("File was apparently not uploaded: [%s]", entry)
                failed = True
            else:
                if size != contents[key]:
                    logger.error("File size differs [%s]: expected %s bytes but got %s bytes", entry, size, contents[key])
                    failed = True

    if not failed:
        logger.info("Completed verifying Amazon S3 bucket contents (no problems found).")
    else:
        logger.error("There were differences between source directory and target S3 bucket.")
        raise ValueError("There were differences between source directory and target S3 bucket.")


#########################################################################
# Main routine
########################################################################

if __name__ == "__main__":
    sys.exit(cli())



# vim: set ft=python ts=4 sw=4 expandtab:
# ruff: noqa: PLC0415
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Cedar Backup, release 3
# Purpose  : Scripts published by UV as part of the Python package
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

"""
Scripts published by UV as part of the Python package.
"""

import sys


def cback3():
    """Implementation of the cback3 script, the main command-line interface."""
    from CedarBackup3.cli import cli

    result = cli()
    sys.exit(result)


def amazons3():
    """Implementation of the cback3-amazons3-sync script."""
    from CedarBackup3.tools.amazons3 import cli

    result = cli()
    sys.exit(result)


def span():
    """Implementation of the cback3-span script."""
    from CedarBackup3.tools.span import cli

    result = cli()
    sys.exit(result)



# vim: set ft=python ts=4 sw=4 expandtab:
# ruff: noqa: S103
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
#              C E D A R
#          S O L U T I O N S       "Software done right."
#           S O F T W A R E
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Copyright (c) 2004-2007,2010,2015 Kenneth J. Pronovici.
# All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# Version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# Copies of the GNU General Public License are available from
# the Free Software Foundation website, http://www.gnu.org/.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Author   : Kenneth J. Pronovici <pronovic@ieee.org>
# Language : Python 3
# Project  : Cedar Backup, release 3
# Purpose  : Tests peer functionality.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

########################################################################
# Module documentation
########################################################################

"""
Unit tests for CedarBackup3/peer.py.

Code Coverage
=============

   This module contains individual tests for most of the public functions and
   classes implemented in peer.py, including the ``LocalPeer`` and ``RemotePeer``
   classes.

   Unfortunately, some of the code can't be tested.  In particular, the stage
   code allows the caller to change ownership on files.  Generally, this can
   only be done by root, and most people won't be running these tests as root.
   As such, we can't test this functionality.  There are also some other pieces
   of functionality that can only be tested in certain environments (see
   below).

Naming Conventions
==================

   I prefer to avoid large unit tests which validate more than one piece of
   functionality, and I prefer to avoid using overly descriptive (read: long)
   test names, as well.  Instead, I use lots of very small tests that each
   validate one specific thing.  These small tests are then named with an index
   number, yielding something like ``testAddDir_001`` or ``testValidate_010``.
   Each method has a docstring describing what it's supposed to accomplish.  I
   feel that this makes it easier to judge how important a given failure is,
   and also makes it somewhat easier to diagnose and fix individual problems.

Full vs. Reduced Tests
======================

   Some Cedar Backup regression tests require a specialized environment in
   order to run successfully.  This environment won't necessarily be available
   on every build system out there (for instance, on a Debian autobuilder).
   Because of this, the default behavior is to run a "reduced feature set" test
   suite that has no surprising system, kernel or network requirements.  If you
   want to run all of the tests, set PEERTESTS_FULL to "Y" in the environment.

   In this module, network-related testing is what causes us our biggest
   problems.  In order to test the RemotePeer, we need a "remote" host that we
   can rcp to and from.  We want to fall back on using localhost and the
   current user, but that might not be safe or appropriate.  As such, we'll
   only run these tests if PEERTESTS_FULL is set to "Y" in the environment.

@author Kenneth J. Pronovici <pronovic@ieee.org>
"""


########################################################################
# Import modules and do runtime validations
########################################################################

# Import standard modules
import os
import stat
import tempfile
import unittest

from CedarBackup3.peer import DEF_COLLECT_INDICATOR, DEF_RCP_COMMAND, DEF_RSH_COMMAND, DEF_STAGE_INDICATOR, LocalPeer, RemotePeer
from CedarBackup3.testutil import (
    buildPath,
    configureLogging,
    extractTar,
    failUnlessAssignRaises,
    findResources,
    getLogin,
    getMaskAsMode,
    platformWindows,
    removedir,
)
from CedarBackup3.util import isRunningAsRoot, pathJoin

#######################################################################
# Module-wide configuration and constants
#######################################################################

DATA_DIRS = [
    "./data",
    "./tests/data",
]
RESOURCES = [
    "tree1.tar.gz",
    "tree2.tar.gz",
    "tree9.tar.gz",
]

REMOTE_HOST = "localhost"  # Always use login@localhost as our "remote" host
NONEXISTENT_FILE = "bogus"  # This file name should never exist
NONEXISTENT_HOST = "hostname.invalid"  # RFC 2606 reserves the ".invalid" TLD for "obviously invalid" names
NONEXISTENT_USER = "unittestuser"  # This user name should never exist on localhost
NONEXISTENT_CMD = "/bogus/~~~ZZZZ/bad/not/there"  # This command should never exist in the filesystem

SAFE_RCP_COMMAND = "/usr/bin/scp -O -B -q -C -o ConnectTimeout=1"  # set a connection timeout so invalid hosts don't hang
SAFE_RSH_COMMAND = "/usr/bin/ssh -o ConnectTimeout=1"  # set a connection timeout so invalid hosts don't hang


#######################################################################
# Utility functions
#######################################################################


def runAllTests():
    """Returns true/false depending on whether the full test suite should be run."""
    if "PEERTESTS_FULL" in os.environ:
        return os.environ["PEERTESTS_FULL"] == "Y"
    else:
        return False


#######################################################################
# Test Case Classes
#######################################################################

######################
# TestLocalPeer class
######################


class TestLocalPeer(unittest.TestCase):
    """Tests for the LocalPeer class."""

    ################
    # Setup methods
    ################

    @classmethod
    def setUpClass(cls):
        configureLogging()

    def setUp(self):
        try:
            self.tmpdir = tempfile.mkdtemp()
            self.resources = findResources(RESOURCES, DATA_DIRS)
        except Exception as e:
            self.fail(e)

    def tearDown(self):
        try:
            removedir(self.tmpdir)
        except:
            pass

    ##################
    # Utility methods
    ##################

    def extractTar(self, tarname):
        """Extracts a tarfile with a particular name."""
        extractTar(self.tmpdir, self.resources["%s.tar.gz" % tarname])

    def buildPath(self, components):
        """Builds a complete search path from a list of components."""
        components.insert(0, self.tmpdir)
        return buildPath(components)

    def getFileMode(self, components):
        """Calls buildPath on components and then returns file mode for the file."""
        return stat.S_IMODE(os.stat(self.buildPath(components)).st_mode)

    def failUnlessAssignRaises(self, exception, obj, prop, value):
        """Equivalent of :any:`failUnlessRaises`, but used for property assignments instead."""
        failUnlessAssignRaises(self, exception, obj, prop, value)

    ###########################
    # Test basic functionality
    ###########################

    def testBasic_001(self):
        """
        Make sure exception is thrown for non-absolute collect directory.
        """
        name = "peer1"
        collectDir = "whatever/something/else/not/absolute"
        self.assertRaises(ValueError, LocalPeer, name, collectDir)

    def testBasic_002(self):
        """
        Make sure attributes are set properly for valid constructor input.
        """
        name = "peer1"
        collectDir = "/absolute/path/name"
        ignoreFailureMode = "all"
        peer = LocalPeer(name, collectDir, ignoreFailureMode)
        self.assertEqual(name, peer.name)
        self.assertEqual(collectDir, peer.collectDir)
        self.assertEqual(ignoreFailureMode, peer.ignoreFailureMode)

    def testBasic_003(self):
        """
        Make sure attributes are set properly for valid constructor input, with
        spaces in the collect directory path.
        """
        name = "peer1"
        collectDir = "/ absolute / path/   name "
        peer = LocalPeer(name, collectDir)
        self.assertEqual(name, peer.name)
        self.assertEqual(collectDir, peer.collectDir)

    def testBasic_004(self):
        """
        Make sure assignment works for all valid failure modes.
        """
        name = "peer1"
        collectDir = "/absolute/path/name"
        ignoreFailureMode = "all"
        peer = LocalPeer(name, collectDir, ignoreFailureMode)
        self.assertEqual("all", peer.ignoreFailureMode)
        peer.ignoreFailureMode = "none"
        self.assertEqual("none", peer.ignoreFailureMode)
        peer.ignoreFailureMode = "daily"
        self.assertEqual("daily", peer.ignoreFailureMode)
        peer.ignoreFailureMode = "weekly"
        self.assertEqual("weekly", peer.ignoreFailureMode)
        self.failUnlessAssignRaises(ValueError, peer, "ignoreFailureMode", "bogus")

    ###############################
    # Test checkCollectIndicator()
    ###############################

    def testCheckCollectIndicator_001(self):
        """
        Attempt to check collect indicator with non-existent collect directory.
        """
        name = "peer1"
        collectDir = self.buildPath([NONEXISTENT_FILE])
        self.assertTrue(not os.path.exists(collectDir))
        peer = LocalPeer(name, collectDir)
        result = peer.checkCollectIndicator()
        self.assertEqual(False, result)

    def testCheckCollectIndicator_002(self):
        """
        Attempt to check collect indicator with non-readable collect directory.
        """
        name = "peer1"
        collectDir = self.buildPath(["collect"])
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        os.chmod(collectDir, 0o200)  # user can't read his own directory
        peer = LocalPeer(name, collectDir)
        result = peer.checkCollectIndicator()
        self.assertEqual(False, result)
        os.chmod(collectDir, 0o777)  # so we can remove it safely

    def testCheckCollectIndicator_003(self):
        """
        Attempt to check collect indicator collect indicator file that does not exist.
        """
        name = "peer1"
        collectDir = self.buildPath(["collect"])
        collectIndicator = self.buildPath(["collect", DEF_COLLECT_INDICATOR])
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(not os.path.exists(collectIndicator))
        peer = LocalPeer(name, collectDir)
        result = peer.checkCollectIndicator()
        self.assertEqual(False, result)

    def testCheckCollectIndicator_004(self):
        """
        Attempt to check collect indicator collect indicator file that does not exist, custom name.
        """
        name = "peer1"
        collectDir = self.buildPath(["collect"])
        collectIndicator = self.buildPath(["collect", NONEXISTENT_FILE])
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(not os.path.exists(collectIndicator))
        peer = LocalPeer(name, collectDir)
        result = peer.checkCollectIndicator(collectIndicator=NONEXISTENT_FILE)
        self.assertEqual(False, result)

    def testCheckCollectIndicator_005(self):
        """
        Attempt to check collect indicator collect indicator file that does exist.
        """
        name = "peer1"
        collectDir = self.buildPath(["collect"])
        collectIndicator = self.buildPath(["collect", DEF_COLLECT_INDICATOR])
        os.mkdir(collectDir)
        with open(collectIndicator, "w") as f:
            f.write("")  # touch the file
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(collectIndicator))
        peer = LocalPeer(name, collectDir)
        result = peer.checkCollectIndicator()
        self.assertEqual(True, result)

    def testCheckCollectIndicator_006(self):
        """
        Attempt to check collect indicator collect indicator file that does exist, custom name.
        """
        name = "peer1"
        collectDir = self.buildPath(["collect"])
        collectIndicator = self.buildPath(["collect", "different"])
        os.mkdir(collectDir)
        with open(collectIndicator, "w") as f:
            f.write("")  # touch the file
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(collectIndicator))
        peer = LocalPeer(name, collectDir)
        result = peer.checkCollectIndicator(collectIndicator="different")
        self.assertEqual(True, result)

    def testCheckCollectIndicator_007(self):
        """
        Attempt to check collect indicator collect indicator file that does exist,
        with spaces in the collect directory path.
        """
        name = "peer1"
        collectDir = self.buildPath(["collect directory here"])
        collectIndicator = self.buildPath(["collect directory here", DEF_COLLECT_INDICATOR])
        os.mkdir(collectDir)
        with open(collectIndicator, "w") as f:
            f.write("")  # touch the file
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(collectIndicator))
        peer = LocalPeer(name, collectDir)
        result = peer.checkCollectIndicator()
        self.assertEqual(True, result)

    @unittest.skipIf(platformWindows(), "Behavior differs on Windows")
    def testCheckCollectIndicator_008(self):
        """
        Attempt to check collect indicator collect indicator file that does exist, custom name,
        with spaces in the collect directory path and collect indicator file name.
        """
        name = "peer1"
        collectDir = self.buildPath([" collect dir "])
        collectIndicator = self.buildPath([" collect dir ", "different, file"])
        os.mkdir(collectDir)
        with open(collectIndicator, "w") as f:
            f.write("")  # touch the file
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(collectIndicator))
        peer = LocalPeer(name, collectDir)
        result = peer.checkCollectIndicator(collectIndicator="different, file")
        self.assertEqual(True, result)

    #############################
    # Test writeStageIndicator()
    #############################

    def testWriteStageIndicator_001(self):
        """
        Attempt to write stage indicator with non-existent collect directory.
        """
        name = "peer1"
        collectDir = self.buildPath([NONEXISTENT_FILE])
        self.assertTrue(not os.path.exists(collectDir))
        peer = LocalPeer(name, collectDir)
        self.assertRaises(ValueError, peer.writeStageIndicator)

    @unittest.skipIf(platformWindows(), "Behavior differs on Windows")
    def testWriteStageIndicator_002(self):
        """
        Attempt to write stage indicator with non-writable collect directory.
        """
        if not isRunningAsRoot():  # root doesn't get this error
            name = "peer1"
            collectDir = self.buildPath(["collect"])
            os.mkdir(collectDir)
            self.assertTrue(os.path.exists(collectDir))
            os.chmod(collectDir, 0o500)  # read-only for user
            peer = LocalPeer(name, collectDir)
            self.assertRaises((IOError, OSError), peer.writeStageIndicator)
            os.chmod(collectDir, 0o777)  # so we can remove it safely

    @unittest.skipIf(platformWindows(), "Behavior differs on Windows")
    def testWriteStageIndicator_003(self):
        """
        Attempt to write stage indicator with non-writable collect directory, custom name.
        """
        if not isRunningAsRoot():  # root doesn't get this error
            name = "peer1"
            collectDir = self.buildPath(["collect"])
            os.mkdir(collectDir)
            self.assertTrue(os.path.exists(collectDir))
            os.chmod(collectDir, 0o500)  # read-only for user
            peer = LocalPeer(name, collectDir)
            self.assertRaises((IOError, OSError), peer.writeStageIndicator, stageIndicator="something")
            os.chmod(collectDir, 0o777)  # so we can remove it safely

    def testWriteStageIndicator_004(self):
        """
        Attempt to write stage indicator in a valid directory.
        """
        name = "peer1"
        collectDir = self.buildPath(["collect"])
        stageIndicator = self.buildPath(["collect", DEF_STAGE_INDICATOR])
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        peer = LocalPeer(name, collectDir)
        peer.writeStageIndicator()
        self.assertTrue(os.path.exists(stageIndicator))

    def testWriteStageIndicator_005(self):
        """
        Attempt to write stage indicator in a valid directory, custom name.
        """
        name = "peer1"
        collectDir = self.buildPath(["collect"])
        stageIndicator = self.buildPath(["collect", "whatever"])
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        peer = LocalPeer(name, collectDir)
        peer.writeStageIndicator(stageIndicator="whatever")
        self.assertTrue(os.path.exists(stageIndicator))

    def testWriteStageIndicator_006(self):
        """
        Attempt to write stage indicator in a valid directory, with spaces
        in the directory name.
        """
        name = "peer1"
        collectDir = self.buildPath(["collect from this directory"])
        stageIndicator = self.buildPath(["collect from this directory", DEF_STAGE_INDICATOR])
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        peer = LocalPeer(name, collectDir)
        peer.writeStageIndicator()
        self.assertTrue(os.path.exists(stageIndicator))

    def testWriteStageIndicator_007(self):
        """
        Attempt to write stage indicator in a valid directory, custom name,
        with spaces in the directory name and the file name.
        """
        name = "peer1"
        collectDir = self.buildPath(["collect ME"])
        stageIndicator = self.buildPath(["collect ME", "   whatever-it-takes you"])
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        peer = LocalPeer(name, collectDir)
        peer.writeStageIndicator(stageIndicator="   whatever-it-takes you")
        self.assertTrue(os.path.exists(stageIndicator))

    ###################
    # Test stagePeer()
    ###################

    def testStagePeer_001(self):
        """
        Attempt to stage files with non-existent collect directory.
        """
        name = "peer1"
        collectDir = self.buildPath([NONEXISTENT_FILE])
        targetDir = self.buildPath(["target"])
        os.mkdir(targetDir)
        self.assertTrue(not os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        peer = LocalPeer(name, collectDir)
        self.assertRaises(ValueError, peer.stagePeer, targetDir=targetDir)

    def testStagePeer_002(self):
        """
        Attempt to stage files with non-readable collect directory.
        """
        name = "peer1"
        collectDir = self.buildPath(["collect"])
        targetDir = self.buildPath(["target"])
        os.mkdir(collectDir)
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        os.chmod(collectDir, 0o200)  # user can't read his own directory
        peer = LocalPeer(name, collectDir)
        self.assertRaises((IOError, OSError), peer.stagePeer, targetDir=targetDir)
        os.chmod(collectDir, 0o777)  # so we can remove it safely

    def testStagePeer_003(self):
        """
        Attempt to stage files with non-absolute target directory.
        """
        name = "peer1"
        collectDir = self.buildPath(["collect"])
        targetDir = "this/is/not/absolute"
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        peer = LocalPeer(name, collectDir)
        self.assertRaises(ValueError, peer.stagePeer, targetDir=targetDir)

    def testStagePeer_004(self):
        """
        Attempt to stage files with non-existent target directory.
        """
        name = "peer1"
        collectDir = self.buildPath(["collect"])
        targetDir = self.buildPath(["target"])
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(not os.path.exists(targetDir))
        peer = LocalPeer(name, collectDir)
        self.assertRaises(ValueError, peer.stagePeer, targetDir=targetDir)

    @unittest.skipIf(platformWindows(), "Behavior differs on Windows")
    def testStagePeer_005(self):
        """
        Attempt to stage files with non-writable target directory.
        """
        if not isRunningAsRoot():  # root doesn't get this error
            self.extractTar("tree1")
            name = "peer1"
            collectDir = self.buildPath(["tree1"])
            targetDir = self.buildPath(["target"])
            os.mkdir(targetDir)
            self.assertTrue(os.path.exists(collectDir))
            self.assertTrue(os.path.exists(targetDir))
            os.chmod(targetDir, 0o500)  # read-only for user
            peer = LocalPeer(name, collectDir)
            self.assertRaises((IOError, OSError), peer.stagePeer, targetDir=targetDir)
            os.chmod(targetDir, 0o777)  # so we can remove it safely
            self.assertEqual(0, len(os.listdir(targetDir)))

    def testStagePeer_006(self):
        """
        Attempt to stage files with empty collect directory.
        *Note:* This test assumes that scp returns an error if the directory is empty.
        """
        self.extractTar("tree2")
        name = "peer1"
        collectDir = self.buildPath(["tree2", "dir001"])
        targetDir = self.buildPath(["target"])
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        peer = LocalPeer(name, collectDir)
        self.assertRaises(IOError, peer.stagePeer, targetDir=targetDir)
        stagedFiles = os.listdir(targetDir)
        self.assertEqual([], stagedFiles)

    @unittest.skipIf(platformWindows(), "Behavior differs on Windows")
    def testStagePeer_007(self):
        """
        Attempt to stage files with empty collect directory, where the target
        directory name contains spaces.
        """
        self.extractTar("tree2")
        name = "peer1"
        collectDir = self.buildPath(["tree2", "dir001"])
        targetDir = self.buildPath([" target directory "])  # windows doesn't like paths that start or end with a space
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        peer = LocalPeer(name, collectDir)
        self.assertRaises(IOError, peer.stagePeer, targetDir=targetDir)
        stagedFiles = os.listdir(targetDir)
        self.assertEqual([], stagedFiles)

    def testStagePeer_008(self):
        """
        Attempt to stage files with non-empty collect directory.
        """
        self.extractTar("tree1")
        name = "peer1"
        collectDir = self.buildPath(["tree1"])
        targetDir = self.buildPath(["target"])
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        self.assertEqual(0, len(os.listdir(targetDir)))
        peer = LocalPeer(name, collectDir)
        count = peer.stagePeer(targetDir=targetDir)
        self.assertEqual(7, count)
        stagedFiles = os.listdir(targetDir)
        self.assertEqual(7, len(stagedFiles))
        self.assertTrue("file001" in stagedFiles)
        self.assertTrue("file002" in stagedFiles)
        self.assertTrue("file003" in stagedFiles)
        self.assertTrue("file004" in stagedFiles)
        self.assertTrue("file005" in stagedFiles)
        self.assertTrue("file006" in stagedFiles)
        self.assertTrue("file007" in stagedFiles)

    def testStagePeer_009(self):
        """
        Attempt to stage files with non-empty collect directory, where the
        target directory name contains spaces.
        """
        self.extractTar("tree1")
        name = "peer1"
        collectDir = self.buildPath(["tree1"])
        targetDir = self.buildPath(["target directory place"])
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        self.assertEqual(0, len(os.listdir(targetDir)))
        peer = LocalPeer(name, collectDir)
        count = peer.stagePeer(targetDir=targetDir)
        self.assertEqual(7, count)
        stagedFiles = os.listdir(targetDir)
        self.assertEqual(7, len(stagedFiles))
        self.assertTrue("file001" in stagedFiles)
        self.assertTrue("file002" in stagedFiles)
        self.assertTrue("file003" in stagedFiles)
        self.assertTrue("file004" in stagedFiles)
        self.assertTrue("file005" in stagedFiles)
        self.assertTrue("file006" in stagedFiles)
        self.assertTrue("file007" in stagedFiles)

    def testStagePeer_010(self):
        """
        Attempt to stage files with non-empty collect directory containing links and directories.
        """
        self.extractTar("tree9")
        name = "peer1"
        collectDir = self.buildPath(["tree9"])
        targetDir = self.buildPath(["target"])
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        self.assertEqual(0, len(os.listdir(targetDir)))
        peer = LocalPeer(name, collectDir)
        self.assertRaises(ValueError, peer.stagePeer, targetDir=targetDir)

    @unittest.skipIf(platformWindows(), "Behavior differs on Windows")
    def testStagePeer_011(self):
        """
        Attempt to stage files with non-empty collect directory and attempt to set valid permissions.
        """
        self.extractTar("tree1")
        name = "peer1"
        collectDir = self.buildPath(["tree1"])
        targetDir = self.buildPath(["target"])
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        self.assertEqual(0, len(os.listdir(targetDir)))
        peer = LocalPeer(name, collectDir)
        if getMaskAsMode() == 0o400:
            permissions = 0o642  # arbitrary, but different than umask would give
        else:
            permissions = 0o400  # arbitrary
        count = peer.stagePeer(targetDir=targetDir, permissions=permissions)
        self.assertEqual(7, count)
        stagedFiles = os.listdir(targetDir)
        self.assertEqual(7, len(stagedFiles))
        self.assertTrue("file001" in stagedFiles)
        self.assertTrue("file002" in stagedFiles)
        self.assertTrue("file003" in stagedFiles)
        self.assertTrue("file004" in stagedFiles)
        self.assertTrue("file005" in stagedFiles)
        self.assertTrue("file006" in stagedFiles)
        self.assertTrue("file007" in stagedFiles)
        self.assertEqual(permissions, self.getFileMode(["target", "file001"]))
        self.assertEqual(permissions, self.getFileMode(["target", "file002"]))
        self.assertEqual(permissions, self.getFileMode(["target", "file003"]))
        self.assertEqual(permissions, self.getFileMode(["target", "file004"]))
        self.assertEqual(permissions, self.getFileMode(["target", "file005"]))
        self.assertEqual(permissions, self.getFileMode(["target", "file006"]))
        self.assertEqual(permissions, self.getFileMode(["target", "file007"]))


######################
# TestRemotePeer class
######################


class TestRemotePeer(unittest.TestCase):
    """Tests for the RemotePeer class."""

    ################
    # Setup methods
    ################

    @classmethod
    def setUpClass(cls):
        configureLogging()

        # Full tests require configured SSH connectivity, so I want to use my own
        # scp/ssh scripts that are set up to use an SSH agent.  This means other
        # people won't be able to run the full tests... but that's always been
        # true, since I don't document the system setup needed to make them work.
        if runAllTests():
            from CedarBackup3.util import PathResolverSingleton

            mapping = {}
            mapping["/usr/bin/ssh"] = pathJoin(os.environ["HOME"], "util", "ssh")
            mapping["/usr/bin/scp"] = pathJoin(os.environ["HOME"], "util", "scp")
            singleton = PathResolverSingleton.getInstance()
            singleton.fill(mapping)

    def setUp(self):
        try:
            self.tmpdir = tempfile.mkdtemp()
            self.resources = findResources(RESOURCES, DATA_DIRS)
        except Exception as e:
            self.fail(e)

    def tearDown(self):
        try:
            removedir(self.tmpdir)
        except:
            pass

    ##################
    # Utility methods
    ##################

    def extractTar(self, tarname):
        """Extracts a tarfile with a particular name."""
        extractTar(self.tmpdir, self.resources["%s.tar.gz" % tarname])

    def buildPath(self, components):
        """Builds a complete search path from a list of components."""
        components.insert(0, self.tmpdir)
        return buildPath(components)

    def getFileMode(self, components):
        """Calls buildPath on components and then returns file mode for the file."""
        return stat.S_IMODE(os.stat(self.buildPath(components)).st_mode)

    def failUnlessAssignRaises(self, exception, obj, prop, value):
        """Equivalent of :any:`failUnlessRaises`, but used for property assignments instead."""
        failUnlessAssignRaises(self, exception, obj, prop, value)

    ############################
    # Tests basic functionality
    ############################

    def testBasic_001(self):
        """
        Make sure exception is thrown for non-absolute collect or working directory.
        """
        name = REMOTE_HOST
        collectDir = "whatever/something/else/not/absolute"
        workingDir = "/tmp"
        remoteUser = getLogin()
        self.assertRaises(ValueError, RemotePeer, name, collectDir, workingDir, remoteUser)

        name = REMOTE_HOST
        collectDir = "/whatever/something/else/not/absolute"
        workingDir = "tmp"
        remoteUser = getLogin()
        self.assertRaises(ValueError, RemotePeer, name, collectDir, workingDir, remoteUser)

    def testBasic_002(self):
        """
        Make sure attributes are set properly for valid constructor input.
        """
        name = REMOTE_HOST
        collectDir = "/absolute/path/name"
        workingDir = "/tmp"
        remoteUser = getLogin()
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        self.assertEqual(name, peer.name)
        self.assertEqual(collectDir, peer.collectDir)
        self.assertEqual(workingDir, peer.workingDir)
        self.assertEqual(remoteUser, peer.remoteUser)
        self.assertEqual(None, peer.localUser)
        self.assertEqual(None, peer.rcpCommand)
        self.assertEqual(None, peer.rshCommand)
        self.assertEqual(None, peer.cbackCommand)
        self.assertEqual(DEF_RCP_COMMAND, peer._rcpCommandList)
        self.assertEqual(DEF_RSH_COMMAND, peer._rshCommandList)
        self.assertEqual(None, peer.ignoreFailureMode)

    def testBasic_003(self):
        """
        Make sure attributes are set properly for valid constructor input, where
        the collect directory contains spaces.
        """
        name = REMOTE_HOST
        collectDir = "/absolute/path/to/ a large directory"
        workingDir = "/tmp"
        remoteUser = getLogin()
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        self.assertEqual(name, peer.name)
        self.assertEqual(collectDir, peer.collectDir)
        self.assertEqual(workingDir, peer.workingDir)
        self.assertEqual(remoteUser, peer.remoteUser)
        self.assertEqual(None, peer.localUser)
        self.assertEqual(None, peer.rcpCommand)
        self.assertEqual(None, peer.rshCommand)
        self.assertEqual(None, peer.cbackCommand)
        self.assertEqual(DEF_RCP_COMMAND, peer._rcpCommandList)
        self.assertEqual(DEF_RSH_COMMAND, peer._rshCommandList)

    def testBasic_004(self):
        """
        Make sure attributes are set properly for valid constructor input, custom rcp command.
        """
        name = REMOTE_HOST
        collectDir = "/absolute/path/name"
        workingDir = "/tmp"
        remoteUser = getLogin()
        rcpCommand = "rcp -one --two three \"four five\" 'six seven' eight"
        peer = RemotePeer(name, collectDir, workingDir, remoteUser, rcpCommand)
        self.assertEqual(name, peer.name)
        self.assertEqual(collectDir, peer.collectDir)
        self.assertEqual(workingDir, peer.workingDir)
        self.assertEqual(remoteUser, peer.remoteUser)
        self.assertEqual(None, peer.localUser)
        self.assertEqual(rcpCommand, peer.rcpCommand)
        self.assertEqual(None, peer.rshCommand)
        self.assertEqual(None, peer.cbackCommand)
        self.assertEqual(["rcp", "-one", "--two", "three", "four five", "'six", "seven'", "eight"], peer._rcpCommandList)
        self.assertEqual(DEF_RSH_COMMAND, peer._rshCommandList)

    def testBasic_005(self):
        """
        Make sure attributes are set properly for valid constructor input, custom local user command.
        """
        name = REMOTE_HOST
        collectDir = "/absolute/path/to/ a large directory"
        workingDir = "/tmp"
        remoteUser = getLogin()
        localUser = "pronovic"
        peer = RemotePeer(name, collectDir, workingDir, remoteUser, localUser=localUser)
        self.assertEqual(name, peer.name)
        self.assertEqual(collectDir, peer.collectDir)
        self.assertEqual(workingDir, peer.workingDir)
        self.assertEqual(remoteUser, peer.remoteUser)
        self.assertEqual(localUser, peer.localUser)
        self.assertEqual(None, peer.rcpCommand)
        self.assertEqual(DEF_RCP_COMMAND, peer._rcpCommandList)
        self.assertEqual(DEF_RSH_COMMAND, peer._rshCommandList)

    def testBasic_006(self):
        """
        Make sure attributes are set properly for valid constructor input, custom rsh command.
        """
        name = REMOTE_HOST
        remoteUser = getLogin()
        rshCommand = 'rsh --whatever -something "a b" else'
        peer = RemotePeer(name, remoteUser=remoteUser, rshCommand=rshCommand)
        self.assertEqual(name, peer.name)
        self.assertEqual(None, peer.collectDir)
        self.assertEqual(None, peer.workingDir)
        self.assertEqual(remoteUser, peer.remoteUser)
        self.assertEqual(None, peer.localUser)
        self.assertEqual(None, peer.rcpCommand)
        self.assertEqual(rshCommand, peer.rshCommand)
        self.assertEqual(None, peer.cbackCommand)
        self.assertEqual(DEF_RCP_COMMAND, peer._rcpCommandList)
        self.assertEqual(DEF_RCP_COMMAND, peer._rcpCommandList)
        self.assertEqual(["rsh", "--whatever", "-something", "a b", "else"], peer._rshCommandList)

    def testBasic_007(self):
        """
        Make sure attributes are set properly for valid constructor input, custom cback command.
        """
        name = REMOTE_HOST
        remoteUser = getLogin()
        cbackCommand = "cback --config=whatever --logfile=whatever --mode=064"
        peer = RemotePeer(name, remoteUser=remoteUser, cbackCommand=cbackCommand)
        self.assertEqual(name, peer.name)
        self.assertEqual(None, peer.collectDir)
        self.assertEqual(None, peer.workingDir)
        self.assertEqual(remoteUser, peer.remoteUser)
        self.assertEqual(None, peer.localUser)
        self.assertEqual(None, peer.rcpCommand)
        self.assertEqual(None, peer.rshCommand)
        self.assertEqual(cbackCommand, peer.cbackCommand)

    def testBasic_008(self):
        """
        Make sure assignment works for all valid failure modes.
        """
        peer = RemotePeer(name="name", remoteUser="user", ignoreFailureMode="all")
        self.assertEqual("all", peer.ignoreFailureMode)
        peer.ignoreFailureMode = "none"
        self.assertEqual("none", peer.ignoreFailureMode)
        peer.ignoreFailureMode = "daily"
        self.assertEqual("daily", peer.ignoreFailureMode)
        peer.ignoreFailureMode = "weekly"
        self.assertEqual("weekly", peer.ignoreFailureMode)
        self.failUnlessAssignRaises(ValueError, peer, "ignoreFailureMode", "bogus")

    ###############################
    # Test checkCollectIndicator()
    ###############################

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testCheckCollectIndicator_001(self):
        """
        Attempt to check collect indicator with invalid hostname.
        """
        name = NONEXISTENT_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        remoteUser = getLogin()
        peer = RemotePeer(name, collectDir, workingDir, remoteUser, rcpCommand=SAFE_RCP_COMMAND, rshCommand=SAFE_RSH_COMMAND)
        result = peer.checkCollectIndicator()
        self.assertEqual(False, result)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testCheckCollectIndicator_002(self):
        """
        Attempt to check collect indicator with invalid remote user.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        remoteUser = NONEXISTENT_USER
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser, rcpCommand=SAFE_RCP_COMMAND, rshCommand=SAFE_RSH_COMMAND)
        result = peer.checkCollectIndicator()
        self.assertEqual(False, result)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testCheckCollectIndicator_003(self):
        """
        Attempt to check collect indicator with invalid rcp command.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        remoteUser = getLogin()
        rcpCommand = NONEXISTENT_CMD
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser, rcpCommand)
        result = peer.checkCollectIndicator()
        self.assertEqual(False, result)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testCheckCollectIndicator_004(self):
        """
        Attempt to check collect indicator with non-existent collect directory.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        remoteUser = getLogin()
        self.assertTrue(not os.path.exists(collectDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        result = peer.checkCollectIndicator()
        self.assertEqual(False, result)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testCheckCollectIndicator_005(self):
        """
        Attempt to check collect indicator with non-readable collect directory.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        os.chmod(collectDir, 0o200)  # user can't read his own directory
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        result = peer.checkCollectIndicator()
        self.assertEqual(False, result)
        os.chmod(collectDir, 0o777)  # so we can remove it safely

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testCheckCollectIndicator_006(self):
        """
        Attempt to check collect indicator collect indicator file that does not exist.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        collectIndicator = self.buildPath(["collect", DEF_COLLECT_INDICATOR])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(not os.path.exists(collectIndicator))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        result = peer.checkCollectIndicator()
        self.assertEqual(False, result)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testCheckCollectIndicator_007(self):
        """
        Attempt to check collect indicator collect indicator file that does not exist, custom name.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        collectIndicator = self.buildPath(["collect", NONEXISTENT_FILE])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(not os.path.exists(collectIndicator))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        result = peer.checkCollectIndicator()
        self.assertEqual(False, result)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testCheckCollectIndicator_008(self):
        """
        Attempt to check collect indicator collect indicator file that does not
        exist, where the collect directory contains spaces.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect directory path"])
        workingDir = "/tmp"
        collectIndicator = self.buildPath(["collect directory path", DEF_COLLECT_INDICATOR])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(not os.path.exists(collectIndicator))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        result = peer.checkCollectIndicator()
        self.assertEqual(False, result)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testCheckCollectIndicator_009(self):
        """
        Attempt to check collect indicator collect indicator file that does not
        exist, custom name, where the collect directory contains spaces.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["  you collect here   "])
        workingDir = "/tmp"
        collectIndicator = self.buildPath(["  you collect here   ", NONEXISTENT_FILE])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(not os.path.exists(collectIndicator))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        result = peer.checkCollectIndicator()
        self.assertEqual(False, result)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testCheckCollectIndicator_010(self):
        """
        Attempt to check collect indicator collect indicator file that does exist.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        collectIndicator = self.buildPath(["collect", DEF_COLLECT_INDICATOR])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        with open(collectIndicator, "w") as f:
            f.write("")  # touch the file
        self.assertTrue(os.path.exists(collectIndicator))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        result = peer.checkCollectIndicator()
        self.assertEqual(True, result)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testCheckCollectIndicator_011(self):
        """
        Attempt to check collect indicator collect indicator file that does exist, custom name.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        collectIndicator = self.buildPath(["collect", "whatever"])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        with open(collectIndicator, "w") as f:
            f.write("")  # touch the file
        self.assertTrue(os.path.exists(collectIndicator))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        result = peer.checkCollectIndicator(collectIndicator="whatever")
        self.assertEqual(True, result)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testCheckCollectIndicator_012(self):
        """
        Attempt to check collect indicator collect indicator file that does exist,
        where the collect directory contains spaces.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect NOT"])
        workingDir = "/tmp"
        collectIndicator = self.buildPath(["collect NOT", DEF_COLLECT_INDICATOR])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        with open(collectIndicator, "w") as f:
            f.write("")  # touch the file
        self.assertTrue(os.path.exists(collectIndicator))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        result = peer.checkCollectIndicator()
        self.assertEqual(True, result)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testCheckCollectIndicator_013(self):
        """
        Attempt to check collect indicator collect indicator file that does
        exist, custom name, where the collect directory and indicator file
        contain spaces.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath([" from here collect!"])
        workingDir = "/tmp"
        collectIndicator = self.buildPath([" from here collect!", "whatever, dude"])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        with open(collectIndicator, "w") as f:
            f.write("")  # touch the file
        self.assertTrue(os.path.exists(collectIndicator))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        result = peer.checkCollectIndicator(collectIndicator="whatever, dude")
        self.assertEqual(True, result)

    #############################
    # Test writeStageIndicator()
    #############################

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testWriteStageIndicator_001(self):
        """
        Attempt to write stage indicator with invalid hostname.
        """
        name = NONEXISTENT_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        remoteUser = getLogin()
        peer = RemotePeer(name, collectDir, workingDir, remoteUser, rcpCommand=SAFE_RCP_COMMAND, rshCommand=SAFE_RSH_COMMAND)
        self.assertRaises((IOError, OSError), peer.writeStageIndicator)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testWriteStageIndicator_002(self):
        """
        Attempt to write stage indicator with invalid remote user.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        remoteUser = NONEXISTENT_USER
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser, rcpCommand=SAFE_RCP_COMMAND, rshCommand=SAFE_RSH_COMMAND)
        self.assertRaises((IOError, OSError), peer.writeStageIndicator)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testWriteStageIndicator_003(self):
        """
        Attempt to write stage indicator with invalid rcp command.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        remoteUser = getLogin()
        rcpCommand = NONEXISTENT_CMD
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser, rcpCommand)
        self.assertRaises((IOError, OSError), peer.writeStageIndicator)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testWriteStageIndicator_004(self):
        """
        Attempt to write stage indicator with non-existent collect directory.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        remoteUser = getLogin()
        self.assertTrue(not os.path.exists(collectDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        self.assertRaises(IOError, peer.writeStageIndicator)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testWriteStageIndicator_005(self):
        """
        Attempt to write stage indicator with non-writable collect directory.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        stageIndicator = self.buildPath(["collect", DEF_STAGE_INDICATOR])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(not os.path.exists(stageIndicator))
        os.chmod(collectDir, 0o400)  # read-only for user
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        self.assertRaises((IOError, OSError), peer.writeStageIndicator)
        self.assertTrue(not os.path.exists(stageIndicator))
        os.chmod(collectDir, 0o777)  # so we can remove it safely

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testWriteStageIndicator_006(self):
        """
        Attempt to write stage indicator in a valid directory.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        stageIndicator = self.buildPath(["collect", DEF_STAGE_INDICATOR])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(not os.path.exists(stageIndicator))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        peer.writeStageIndicator()
        self.assertTrue(os.path.exists(stageIndicator))

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testWriteStageIndicator_007(self):
        """
        Attempt to write stage indicator in a valid directory, custom name.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        stageIndicator = self.buildPath(["collect", "newname"])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(not os.path.exists(stageIndicator))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        peer.writeStageIndicator(stageIndicator="newname")
        self.assertTrue(os.path.exists(stageIndicator))

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testWriteStageIndicator_008(self):
        """
        Attempt to write stage indicator in a valid directory that contains
        spaces.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["with spaces collect"])
        workingDir = "/tmp"
        stageIndicator = self.buildPath(["with spaces collect", DEF_STAGE_INDICATOR])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(not os.path.exists(stageIndicator))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser, rcpCommand=SAFE_RCP_COMMAND, rshCommand=SAFE_RSH_COMMAND)
        peer.writeStageIndicator()
        self.assertTrue(os.path.exists(stageIndicator))

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testWriteStageIndicator_009(self):
        """
        Attempt to write stage indicator in a valid directory, custom name, where
        the collect directory and the custom name contain spaces.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect, soon"])
        workingDir = "/tmp"
        stageIndicator = self.buildPath(["collect, soon", "new name with spaces"])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(not os.path.exists(stageIndicator))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser, rcpCommand=SAFE_RCP_COMMAND, rshCommand=SAFE_RSH_COMMAND)
        peer.writeStageIndicator(stageIndicator="new name with spaces")
        self.assertTrue(os.path.exists(stageIndicator))

    ###################
    # Test stagePeer()
    ###################

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_001(self):
        """
        Attempt to stage files with invalid hostname.
        """
        name = NONEXISTENT_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        targetDir = self.buildPath(["target"])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser, rcpCommand=SAFE_RCP_COMMAND, rshCommand=SAFE_RSH_COMMAND)
        self.assertRaises((IOError, OSError), peer.stagePeer, targetDir=targetDir)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_002(self):
        """
        Attempt to stage files with invalid remote user.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        targetDir = self.buildPath(["target"])
        remoteUser = NONEXISTENT_USER
        os.mkdir(collectDir)
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser, rcpCommand=SAFE_RCP_COMMAND, rshCommand=SAFE_RSH_COMMAND)
        self.assertRaises((IOError, OSError), peer.stagePeer, targetDir=targetDir)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_003(self):
        """
        Attempt to stage files with invalid rcp command.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        targetDir = self.buildPath(["target"])
        remoteUser = getLogin()
        rcpCommand = NONEXISTENT_CMD
        os.mkdir(collectDir)
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser, rcpCommand)
        self.assertRaises((IOError, OSError), peer.stagePeer, targetDir=targetDir)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_004(self):
        """
        Attempt to stage files with non-existent collect directory.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        targetDir = self.buildPath(["target"])
        remoteUser = getLogin()
        os.mkdir(targetDir)
        self.assertTrue(not os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        self.assertRaises((IOError, OSError), peer.stagePeer, targetDir=targetDir)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_005(self):
        """
        Attempt to stage files with non-readable collect directory.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        targetDir = self.buildPath(["target"])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        os.chmod(collectDir, 0o200)  # user can't read his own directory
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        self.assertRaises((IOError, OSError), peer.stagePeer, targetDir=targetDir)
        os.chmod(collectDir, 0o777)  # so we can remove it safely

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_006(self):
        """
        Attempt to stage files with non-absolute target directory.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        targetDir = "non/absolute/target"
        remoteUser = getLogin()
        self.assertTrue(not os.path.exists(collectDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        self.assertRaises(ValueError, peer.stagePeer, targetDir=targetDir)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_007(self):
        """
        Attempt to stage files with non-existent target directory.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        targetDir = self.buildPath(["target"])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(not os.path.exists(targetDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        self.assertRaises(ValueError, peer.stagePeer, targetDir=targetDir)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_008(self):
        """
        Attempt to stage files with non-writable target directory.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        targetDir = self.buildPath(["target"])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        os.chmod(targetDir, 0o400)  # read-only for user
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        self.assertRaises((IOError, OSError), peer.stagePeer, targetDir=targetDir)
        os.chmod(collectDir, 0o777)  # so we can remove it safely
        self.assertEqual(0, len(os.listdir(targetDir)))

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_009(self):
        """
        Attempt to stage files with empty collect directory.
        *Note:* This test assumes that scp returns an error if the directory is empty.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        targetDir = self.buildPath(["target"])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        self.assertRaises((IOError, OSError), peer.stagePeer, targetDir=targetDir)
        stagedFiles = os.listdir(targetDir)
        self.assertEqual([], stagedFiles)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_010(self):
        """
        Attempt to stage files with empty collect directory, with a target
        directory that contains spaces.
        *Note:* This test assumes that scp returns an error if the directory is empty.
        """
        name = REMOTE_HOST
        collectDir = self.buildPath(["collect"])
        workingDir = "/tmp"
        targetDir = self.buildPath(["target DIR"])
        remoteUser = getLogin()
        os.mkdir(collectDir)
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        self.assertRaises((IOError, OSError), peer.stagePeer, targetDir=targetDir)
        stagedFiles = os.listdir(targetDir)
        self.assertEqual([], stagedFiles)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_011(self):
        """
        Attempt to stage files with non-empty collect directory.
        """
        self.extractTar("tree1")
        name = REMOTE_HOST
        collectDir = self.buildPath(["tree1"])
        workingDir = "/tmp"
        targetDir = self.buildPath(["target"])
        remoteUser = getLogin()
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        self.assertEqual(0, len(os.listdir(targetDir)))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        count = peer.stagePeer(targetDir=targetDir)
        self.assertEqual(7, count)
        stagedFiles = os.listdir(targetDir)
        self.assertEqual(7, len(stagedFiles))
        self.assertTrue("file001" in stagedFiles)
        self.assertTrue("file002" in stagedFiles)
        self.assertTrue("file003" in stagedFiles)
        self.assertTrue("file004" in stagedFiles)
        self.assertTrue("file005" in stagedFiles)
        self.assertTrue("file006" in stagedFiles)
        self.assertTrue("file007" in stagedFiles)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_012(self):
        """
        Attempt to stage files with non-empty collect directory, with a target
        directory that contains spaces.
        """
        self.extractTar("tree1")
        name = REMOTE_HOST
        collectDir = self.buildPath(["tree1"])
        workingDir = "/tmp"
        targetDir = self.buildPath(["write the target here, now!"])
        remoteUser = getLogin()
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        self.assertEqual(0, len(os.listdir(targetDir)))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        count = peer.stagePeer(targetDir=targetDir)
        self.assertEqual(7, count)
        stagedFiles = os.listdir(targetDir)
        self.assertEqual(7, len(stagedFiles))
        self.assertTrue("file001" in stagedFiles)
        self.assertTrue("file002" in stagedFiles)
        self.assertTrue("file003" in stagedFiles)
        self.assertTrue("file004" in stagedFiles)
        self.assertTrue("file005" in stagedFiles)
        self.assertTrue("file006" in stagedFiles)
        self.assertTrue("file007" in stagedFiles)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_013(self):
        """
        Attempt to stage files with non-empty collect directory containing links and directories.
        *Note:* We assume that scp copies the files even though it returns an error due to directories.
        """
        self.extractTar("tree9")
        name = REMOTE_HOST
        collectDir = self.buildPath(["tree9"])
        workingDir = "/tmp"
        targetDir = self.buildPath(["target"])
        remoteUser = getLogin()
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        self.assertEqual(0, len(os.listdir(targetDir)))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        self.assertRaises((IOError, OSError), peer.stagePeer, targetDir=targetDir)
        stagedFiles = os.listdir(targetDir)
        self.assertEqual(2, len(stagedFiles))
        self.assertTrue("file001" in stagedFiles)
        self.assertTrue("file002" in stagedFiles)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testStagePeer_014(self):
        """
        Attempt to stage files with non-empty collect directory and attempt to set valid permissions.
        """
        self.extractTar("tree1")
        name = REMOTE_HOST
        collectDir = self.buildPath(["tree1"])
        workingDir = "/tmp"
        targetDir = self.buildPath(["target"])
        remoteUser = getLogin()
        os.mkdir(targetDir)
        self.assertTrue(os.path.exists(collectDir))
        self.assertTrue(os.path.exists(targetDir))
        self.assertEqual(0, len(os.listdir(targetDir)))
        peer = RemotePeer(name, collectDir, workingDir, remoteUser)
        if getMaskAsMode() == 0o400:
            permissions = 0o642  # arbitrary, but different than umask would give
        else:
            permissions = 0o400  # arbitrary
        count = peer.stagePeer(targetDir=targetDir, permissions=permissions)
        self.assertEqual(7, count)
        stagedFiles = os.listdir(targetDir)
        self.assertEqual(7, len(stagedFiles))
        self.assertTrue("file001" in stagedFiles)
        self.assertTrue("file002" in stagedFiles)
        self.assertTrue("file003" in stagedFiles)
        self.assertTrue("file004" in stagedFiles)
        self.assertTrue("file005" in stagedFiles)
        self.assertTrue("file006" in stagedFiles)
        self.assertTrue("file007" in stagedFiles)
        self.assertEqual(permissions, self.getFileMode(["target", "file001"]))
        self.assertEqual(permissions, self.getFileMode(["target", "file002"]))
        self.assertEqual(permissions, self.getFileMode(["target", "file003"]))
        self.assertEqual(permissions, self.getFileMode(["target", "file004"]))
        self.assertEqual(permissions, self.getFileMode(["target", "file005"]))
        self.assertEqual(permissions, self.getFileMode(["target", "file006"]))
        self.assertEqual(permissions, self.getFileMode(["target", "file007"]))

    ##############################
    # Test executeRemoteCommand()
    ##############################

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testExecuteRemoteCommand(self):
        """
        Test that a simple remote command succeeds.
        """
        target = self.buildPath(["test.txt"])
        name = REMOTE_HOST
        remoteUser = getLogin()
        command = "touch %s" % target
        self.assertFalse(os.path.exists(target))
        peer = RemotePeer(name=name, remoteUser=remoteUser)
        peer.executeRemoteCommand(command)
        self.assertTrue(os.path.exists(target))

    ############################
    # Test _buildCbackCommand()
    ############################

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testBuildCbackCommand_001(self):
        """
        Test with None for cbackCommand and action, False for fullBackup.
        """
        self.assertRaises(ValueError, RemotePeer._buildCbackCommand, None, None, False)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testBuildCbackCommand_002(self):
        """
        Test with None for cbackCommand, "collect" for action, False for fullBackup.
        """
        result = RemotePeer._buildCbackCommand(None, "collect", False)
        self.assertEqual("/usr/bin/cback3 collect", result)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testBuildCbackCommand_003(self):
        """
        Test with "cback" for cbackCommand, "collect" for action, False for fullBackup.
        """
        result = RemotePeer._buildCbackCommand("cback", "collect", False)
        self.assertEqual("cback collect", result)

    @unittest.skipUnless(runAllTests(), "Limited test suite")
    def testBuildCbackCommand_004(self):
        """
        Test with "cback" for cbackCommand, "collect" for action, True for fullBackup.
        """
        result = RemotePeer._buildCbackCommand("cback", "collect", True)
        self.assertEqual("cback --full collect", result)
